 /*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Data Resource API
 * 
 * Copyright (C) 2004, Rob Gagnon
 * Rob Gagnon <rgagnon@networkip.net>
 *
 * Contributions by:
 *   Filip Olsson <filip.olsson@telavox.se>
 *   Steven Sokol <ssokol@sokol-associates.com>
 *   Brian Degenhardt <bmd@fourloop.com>
 * 
 * based on code by:
 * Mark Spencer <markster@linux-support.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License.
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "asterisk.h"
#include <asterisk/lock.h>
#include <asterisk/config.h>
#include <asterisk/logger.h>
#include <asterisk/module.h>
#include <asterisk/options.h>
#include <asterisk/channel.h>
#include <asterisk/cli.h>
#include <asterisk/data.h>
#include <asterisk/utils.h>

ASTDATA_DICT_ATTR attribute_dictionary[] = {
	{ASTDATA_ATTR_ACCOUNTCODE,		ASTDATA_TYPE_STRING,		"AccountCode"},
	{ASTDATA_ATTR_ACLIPADDRESS,	ASTDATA_TYPE_IPADDR,		"ACLIPAddress"},
	{ASTDATA_ATTR_ACLNETMASK,		ASTDATA_TYPE_IPADDR,		"ACLNetmask"},
	{ASTDATA_ATTR_ACLSENSE,			ASTDATA_TYPE_INTEGER,	"ACLSense"},
	{ASTDATA_ATTR_ACTION,			ASTDATA_TYPE_INTEGER,	"Action"},
	{ASTDATA_ATTR_AMAFLAGS,			ASTDATA_TYPE_INTEGER,	"AMAFlags"},
	{ASTDATA_ATTR_APPLICATION,		ASTDATA_TYPE_STRING,		"Application"},
	{ASTDATA_ATTR_ARGUMENTS,		ASTDATA_TYPE_STRING,		"Arguments"},
	{ASTDATA_ATTR_AUTHFLAGS,		ASTDATA_TYPE_INTEGER,	"AuthMethodFlags"},
	{ASTDATA_ATTR_CALLERID,			ASTDATA_TYPE_STRING,		"CallerID"},
	{ASTDATA_ATTR_CALLERID_NAME,	ASTDATA_TYPE_STRING,		"CallerIDName"},
	{ASTDATA_ATTR_CALLERID_NUM,	ASTDATA_TYPE_STRING,		"CallerIDNumber"},
	{ASTDATA_ATTR_CALLINGPRES,		ASTDATA_TYPE_INTEGER,	"CallingPresence"},
	{ASTDATA_ATTR_CANREINVITE,		ASTDATA_TYPE_INTEGER,	"CanReinviteFlag"},
	{ASTDATA_ATTR_CODECFLAGS,		ASTDATA_TYPE_INTEGER,	"CodecFlags"},
	{ASTDATA_ATTR_CONTEXT,			ASTDATA_TYPE_STRING,		"Context"},
	{ASTDATA_ATTR_DATASOURCE,		ASTDATA_TYPE_STRING,		"DataSource"},
	{ASTDATA_ATTR_DTMFMODE,			ASTDATA_TYPE_INTEGER,	"DTMFMode"},
	{ASTDATA_ATTR_EMAIL,				ASTDATA_TYPE_STRING,		"EmailAddress"},
	{ASTDATA_ATTR_ERR_MSG,			ASTDATA_TYPE_STRING,		"ErrorMessage"},
	{ASTDATA_ATTR_ERR_NUM,			ASTDATA_TYPE_INTEGER,	"ErrorNumber"},
	{ASTDATA_ATTR_EXPIRY,			ASTDATA_TYPE_INTEGER,	"Expiry"},
	{ASTDATA_ATTR_EXTENSION,		ASTDATA_TYPE_STRING,		"Extension"},
	{ASTDATA_ATTR_FULLCONTACT,		ASTDATA_TYPE_STRING,		"FullContact"},
	{ASTDATA_ATTR_FULLNAME,			ASTDATA_TYPE_STRING,		"FullName"},
	{ASTDATA_ATTR_IPADDRESS,		ASTDATA_TYPE_IPADDR,		"IPAddress"},
	{ASTDATA_ATTR_LANGUAGE,			ASTDATA_TYPE_STRING,		"Language"},
	{ASTDATA_ATTR_LASTNAMEFLAG,	ASTDATA_TYPE_INTEGER,	"LastNameFlag"},
	{ASTDATA_ATTR_MAILBOX,			ASTDATA_TYPE_STRING,		"Mailbox"},
	{ASTDATA_ATTR_MD5PASSWORD,		ASTDATA_TYPE_STRING,		"MD5Password"},
	{ASTDATA_ATTR_NAME,				ASTDATA_TYPE_STRING,		"Name"},
	{ASTDATA_ATTR_NATFLAG,			ASTDATA_TYPE_INTEGER,	"NATFlag"},
	{ASTDATA_ATTR_NETMASK,			ASTDATA_TYPE_IPADDR,		"NetMask"},
	{ASTDATA_ATTR_NEWPASSWORD,		ASTDATA_TYPE_STRING,		"NewPassword"},
	{ASTDATA_ATTR_NOTRANSFERFLAG,	ASTDATA_TYPE_INTEGER,	"NoTransferFlag"},
	{ASTDATA_ATTR_OPTIONS,			ASTDATA_TYPE_STRING,		"Options"},
	{ASTDATA_ATTR_PUBLICKEY,		ASTDATA_TYPE_STRING,		"PublicKey"},
	{ASTDATA_ATTR_PAGER,				ASTDATA_TYPE_STRING,		"Pager"},
	{ASTDATA_ATTR_PASSWORD,			ASTDATA_TYPE_STRING,		"Password"},
	{ASTDATA_ATTR_PORT,				ASTDATA_TYPE_INTEGER,	"Port"},
	{ASTDATA_ATTR_PRIORITY,			ASTDATA_TYPE_INTEGER,	"Priority"},
	{ASTDATA_ATTR_PRIVATEKEY,		ASTDATA_TYPE_STRING,		"PrivateKey"},
	{ASTDATA_ATTR_REGSECONDS,		ASTDATA_TYPE_INTEGER,	"RegSeconds"},
	{ASTDATA_ATTR_USERAGENT,		ASTDATA_TYPE_STRING,		"UserAgent"},
	{ASTDATA_ATTR_USERINPUT,		ASTDATA_TYPE_STRING,		"UserInput"},
	{ASTDATA_ATTR_USERNAME,			ASTDATA_TYPE_STRING,		"Username"},
	{ASTDATA_ATTR_USERTYPE,			ASTDATA_TYPE_INTEGER,	"UserType"},
	{ASTDATA_ATTR_INVALID,			ASTDATA_TYPE_INVALID,	"EndOfList"}
};

int astdatadebug = 0;

/* Query Interface Handler Item and Mutex Lock */
AST_MUTEX_DEFINE_STATIC(ast_data_lock);

static struct ast_data_handleritem {
	char name[20];
	char desc[80];
	ASTDATA_QTYPE  qtype;
	ast_data_query_handler query_handler;
	struct ast_data_handleritem *next;
} *handlers = NULL;

int
ast_data_legacy_load(ASTDATA_DBINFO *db, char *config_file);

int
ast_data_legacy_config(void);

/***************************************************************************/

static int
ast_data_cli_debug(int fd, int argc, char *argv[]) {
	astdatadebug = 1;
	ast_cli(fd,"data debugging enabled\n");
	return(RESULT_SUCCESS);
}

/***************************************************************************/

static int
ast_data_cli_no_debug(int fd, int argc, char *argv[]) {
	astdatadebug = 0;
	ast_cli(fd,"data debugging disabled\n");
	return(RESULT_SUCCESS);
}

/***************************************************************************/

static struct ast_cli_entry ast_data_debug_cli =
{{"data", "debug", NULL},
	ast_data_cli_debug, "Enable data-debugging",
	"Dumps received and transmitted packages to CLI\n"};
	
static struct ast_cli_entry ast_data_no_debug_cli =
{{"data", "no", "debug", NULL},
	ast_data_cli_no_debug, "Disable data-debugging",
	"Disable data-debugging\n"};

/***************************************************************************/

static int
ast_data_cli_show_handlers(int fd, int argc, char *argv[]) {
	struct ast_data_handleritem *hitem;
	ASTDATA_QTYPE qtype = ASTDATA_QTYPE_INVALID;

	if (handlers) {
		ast_cli(fd,"List of Data Resource Handlers:\n");
	} else {
		ast_cli(fd,"There are no registered Data Resource Handlers.\n\n");
		return(RESULT_SUCCESS);
	}
	ast_mutex_lock(&ast_data_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype != qtype) {
			qtype = hitem->qtype;
			ast_cli(fd, "  Handler(s) for %s:\n", ast_data_qtype2str(qtype));
		}
		if (hitem->query_handler) {
			ast_cli(fd, "    '%s' is registered (%s)\n",
				hitem->name, hitem->desc);
		}
		else {
			ast_cli(fd, "    '%s' has reservation. Nothing registered.\n",
				hitem->name);
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&ast_data_lock);
	if (handlers) {
		ast_cli(fd,"=== End of List ===\n\n");
	}
	return(RESULT_SUCCESS);
}

/***************************************************************************/

static struct ast_cli_entry ast_data_show_handlers_cli = 
{ {"show", "data", "handlers", NULL},
	ast_data_cli_show_handlers, "Shows registered Data Resource Handlers",
	"Shows registered Data Resource Handlers and their descriptions\n"};

/***************************************************************************/

int
ast_data_escape_string(char *to, char *from, int fromlen) {
	int x = fromlen;
	int fp = 0;
	int tp = 0;

	while (x > 0) {
		switch (from[fp]) {
			case '\'':
			case '\\':
			case '"':
				to[tp] = '\\';
				tp++;
				break;
		}
		to[tp] = from[fp];
		tp++;
		fp++;
		x--;
	}
	to[tp] = '\0';
	return(strlen(to));
}

/***************************************************************************/

char *
ast_data_rtrim(char *s) {
	char *p = s+strlen(s)-1;

	while (p != s) {
		if (*p != ' ' && *p != '\r' && *p != '\n') {
			break;
		}
		*p = '\0';
		p--;
	}
	return(s);
}

/***************************************************************************/

struct ast_config *
ast_data_read_handler_flags(char *modname, ASTDATA_DBINFO *dbinfo) {
	struct ast_config *cfg = NULL;
	struct ast_variable *var = NULL;

	cfg = ast_load(ASTDATA_CONFIG_FILE);
	if (!cfg) {
		ast_log(LOG_WARNING, "Unable to load config for %s Data Resource: %s\n",
			modname, ASTDATA_CONFIG_FILE);
		return(NULL);
	}

	/* read the global configuration first */
	var = ast_variable_browse(cfg, "global");
	if (!var) {
		/* no global section? */
		ast_log(LOG_WARNING, "No global section in %s\n", ASTDATA_CONFIG_FILE);
		ast_destroy(cfg);
		return(NULL);
	}

	/* check if this module is supposed to be used.. set flags to register
	 * after the database connection is established
	 */
	while (var) {
		if (!strcasecmp(var->name, "sipfriends") && strstr(var->value, modname)) {
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_SIP_FIND_PEER, NULL);
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_SIP_FIND_PEERS_WITH_MAILBOX, NULL);
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_SIP_FIND_USER, NULL);
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_SIP_UPDATE_PEER, NULL);
		}
		if (!strcasecmp(var->name, "iaxfriends") && strstr(var->value, modname)) {
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_IAX_FIND_PEER, NULL);
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_IAX_FIND_USER, NULL);
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_IAX_UPDATE_PEER, NULL);
		}
		if (!strcasecmp(var->name, "extensionlookup") && strstr(var->value, modname)) {
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_EXTENSION_LOOKUP, NULL);
		}
		if (!strcasecmp(var->name, "voicemail") && strstr(var->value, modname)) {
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_VM_FIND_USER, NULL);
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_VM_CHANGE_PASSWORD, NULL);
		}
		if (!strcasecmp(var->name, "directory") && strstr(var->value, modname)) {
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_VM_DIRECTORY, NULL);
		}
		if (!strcasecmp(var->name, "endofcall") && strstr(var->value, modname)) {
			ast_data_add_qinfo(dbinfo, ASTDATA_QTYPE_END_OF_CALL, NULL);
		}

		var = var->next;
	}

	return(cfg);
}

/***************************************************************************/

/* read a config file, looking for the db configuration fields
 * dbuser, dbpass, dbhost, and dbname (for MYSQL)
 * or "dboption" (for PGSQL)
 */
int
ast_data_legacy_load(ASTDATA_DBINFO *db, char *config_file) {
	struct ast_config *cfg = NULL;
	struct ast_variable *var = NULL;

	memset(db, 0, sizeof(ASTDATA_DBINFO));

	cfg = ast_load(config_file);
	if (!cfg) {
		ast_log(LOG_ERROR,"Could not read %s legacy config\n", config_file);
		return(0);
	}

	var = ast_variable_browse(cfg, "general");
	if (!var) {
		/* nothing configured */
		ast_log(LOG_WARNING, "no general section in %s\n", config_file);
		ast_destroy(cfg);
		return(0);
	}

	while (var) {
		if (!strcasecmp(var->name, "dbhost")) {
			strncpy(db->hostname, var->value, sizeof(db->hostname)-1);
		}
		else if (!strcasecmp(var->name, "dbname")) {
			strncpy(db->dbname, var->value, sizeof(db->dbname)-1);
		}
		else if (!strcasecmp(var->name, "dbuser")) {
			strncpy(db->user, var->value, sizeof(db->user)-1);
		}
		else if (!strcasecmp(var->name, "dbpass")) {
			strncpy(db->password, var->value, sizeof(db->password)-1);
		}
		else if (!strcasecmp(var->name, "dboption")) {
			strncpy(db->options, var->value, sizeof(db->options)-1);
		}
		var = var->next;
	}

	ast_destroy(cfg);
	return(-1);
}

/***************************************************************************/

/*
 * Read old db config, and convert to new data.conf format
 */
int
ast_data_legacy_config(void) {
	ASTDATA_DBINFO sip_config;
	ASTDATA_DBINFO iax_config;
	ASTDATA_DBINFO vm_config;
	int res = 0;
	char filename[256] = "";
	FILE *fp = NULL;

	/* First, read the sip.conf, iax.conf, and voicemail.conf files */
	res = ast_data_legacy_load(&sip_config, ASTDATA_LEGACY_SIP_CONFIG);
	if (!res) {
		return(res);
	}
	res = ast_data_legacy_load(&iax_config, ASTDATA_LEGACY_IAX_CONFIG);
	if (!res) {
		return(res);
	}
	res = ast_data_legacy_load(&vm_config, ASTDATA_LEGACY_VM_CONFIG);
	if (!res) {
		return(res);
	}
	
	/* the path/file to data.conf */
	snprintf(filename, sizeof(filename), "%s/%s",
		AST_CONFIG_DIR, ASTDATA_CONFIG_FILE);

	fp = fopen(filename, "w");
	if (!fp) {
		ast_log(LOG_ERROR, "Could not create new %s file\n", ASTDATA_CONFIG_FILE);
		return(0);
	}

	fprintf(fp, "; data.c auto-created legacy config file\n;\n");
	fprintf(fp, "; please see sample data.conf.sample for information\n");
	fprintf(fp, ";\n[Global]\n; ExtensionLookup =\n; Directory =\nEndOfCall =\n");

	/* Now lets see if there is a configuration to make */
	if (sip_config.hostname[0]) {
		/* sip.conf configured for MySQL */
		fprintf(fp, "SIPFriends = mysql\n");
	}
	else {
		fprintf(fp, "; SIPFriends =\n");
	}

	if (iax_config.hostname[0]) {
		/* iax.conf configured for MySQL */
		fprintf(fp, "IAXFriends = mysql\n");
	}
	else if (iax_config.options[0]) {
		/* iax.conf configured for PGSQL */
		fprintf(fp, "IAXFriends = pgsql\n");
	}
	else {
		fprintf(fp, "; IAXFriends =\n");
	}

	if (vm_config.hostname[0]) {
		/* voicemail.conf configured for MySQL */
		fprintf(fp, "VoiceMail = mysql\n");
	}
	else if (vm_config.options[0]) {
		/* voicemail.conf configured for PGSQL */
		fprintf(fp, "VoiceMail = pgsql\n");
	}
	else {
		fprintf(fp, "; VoiceMail =\n");
	}

	fprintf(fp, "\n[mysql]\n");
	fprintf(fp,
		"; If all MySQL lookups use same database, set these options:\n");
	fprintf(fp, "; QType = Default\n");
	fprintf(fp, "; Hostname =\n; DBName =\n; User =\n; Password =\n");
	fprintf(fp, "; Port = 3306\n; Sock = /tmp/mysql.sock\n\n");

	fprintf(fp, "; SIP Specific Database Lookup\n");
	if (sip_config.hostname[0]) {
		fprintf(fp, "QType = SIPFriends\n");
		fprintf(fp, "Hostname = %s\n", sip_config.hostname);
		fprintf(fp, "DBName = %s\n", sip_config.dbname);
		fprintf(fp, "User = %s\n", sip_config.user);
		fprintf(fp, "Password = %s\n", sip_config.password);
		fprintf(fp, "; Port = 3306\n; Sock = /tmp/mysql.sock\n");
	}
	else {
		fprintf(fp, "; QType = SIPFriends\n");
		fprintf(fp, "; Hostname =\n");
		fprintf(fp, "; DBName =\n");
		fprintf(fp, "; User =\n");
		fprintf(fp, "; Password =\n");
		fprintf(fp, "; Port = 3306\n; Sock = /tmp/mysql.sock\n");
	}

	fprintf(fp, "\n; IAX Specific Database Lookup\n");
	if (iax_config.hostname[0]) {
		fprintf(fp, "QType = IAXFriends\n");
		fprintf(fp, "Hostname = %s\n", iax_config.hostname);
		fprintf(fp, "DBName = %s\n", iax_config.dbname);
		fprintf(fp, "User = %s\n", iax_config.user);
		fprintf(fp, "Password = %s\n", iax_config.password);
		fprintf(fp, "; Port = 3306\n; Sock = /tmp/mysql.sock\n");
	}
	else {
		fprintf(fp, "; QType = IAXFriends\n");
		fprintf(fp, "; Hostname =\n");
		fprintf(fp, "; DBName =\n");
		fprintf(fp, "; User =\n");
		fprintf(fp, "; Password =\n");
		fprintf(fp, "; Port = 3306\n; Sock = /tmp/mysql.sock\n");
	}

	fprintf(fp, "\n; Voicemail Specific Database Lookup\n");
	if (vm_config.hostname[0]) {
		fprintf(fp, "QType = VoiceMail\n");
		fprintf(fp, "Hostname = %s\n", vm_config.hostname);
		fprintf(fp, "DBName = %s\n", vm_config.dbname);
		fprintf(fp, "User = %s\n", vm_config.user);
		fprintf(fp, "Password = %s\n", vm_config.password);
		fprintf(fp, "; Port = 3306\n; Sock = /tmp/mysql.sock\n");
	}
	else {
		fprintf(fp, "; QType = VoiceMail\n");
		fprintf(fp, "; Hostname =\n");
		fprintf(fp, "; DBName =\n");
		fprintf(fp, "; User =\n");
		fprintf(fp, "; Password =\n");
		fprintf(fp, "; Port = 3306\n; Sock = /tmp/mysql.sock\n");
	}

	fprintf(fp, "\n[pgsql]\n");
	fprintf(fp,
		"; If all PostGreSQL lookups use same database, set these options:\n");
	fprintf(fp, "; QType = Default\n");
	fprintf(fp, "; DBOptions =\n");

	fprintf(fp, "\n; Voicemail Specific Database Lookup\n");
	if (vm_config.options[0]) {
		fprintf(fp, "QType = VoiceMail\n");
		fprintf(fp, "DBOptions = %s\n", vm_config.options);
	}
	else {
		fprintf(fp, "; QType = VoiceMail\n");
		fprintf(fp, "; DBOptions =\n");
	}

	fprintf(fp, "\n; IAX Specific Database Lookup\n");
	if (iax_config.options[0]) {
		fprintf(fp, "QType = IAXFriends\n");
		fprintf(fp, "DBOptions = %s\n", iax_config.options);
	}
	else {
		fprintf(fp, "; QType = IAXFriends\n");
		fprintf(fp, "; DBOptions =\n");
	}

	fprintf(fp, "\n[odbc]\n");
	fprintf(fp,
		"; If all UnixODBC lookups use same DSN, set these options:\n");
	fprintf(fp, "; QType = Default\n");
	fprintf(fp, "; DSN =\n\n");
	fprintf(fp, "; User =\n");
	fprintf(fp, "; Password =\n");
	fprintf(fp, "\n; End of File\n");
	fclose(fp);
	return(-1);
}

/***************************************************************************/

int
ast_data_load(void) {
	struct ast_config *cfg = NULL;
	struct ast_variable *var = NULL;
	
	ast_cli_register(&ast_data_show_handlers_cli);
	ast_cli_register(&ast_data_no_debug_cli);
	ast_cli_register(&ast_data_debug_cli);

	/* read the global section of the config file */
	cfg = ast_load(ASTDATA_CONFIG_FILE);
	if (!cfg) {
		ast_log(LOG_WARNING, "Unable to load config %s\n", ASTDATA_CONFIG_FILE);
		ast_log(LOG_WARNING, "Performing Legacy config with %s and %s\n",
			ASTDATA_LEGACY_SIP_CONFIG, ASTDATA_LEGACY_VM_CONFIG);
		if (!ast_data_legacy_config()) {
			ast_log(LOG_ERROR, "Legacy config failed!\n");
			return(0);
		}
		/* Legacy config should have built a data.conf file */
		cfg = ast_load(ASTDATA_CONFIG_FILE);
		if (!cfg) {
			ast_log(LOG_ERROR, "Unable to load config %s\n", ASTDATA_CONFIG_FILE);
			return(0);
		}
	}

	var = ast_variable_browse(cfg, "global");
	if (!var) {
		/* nothing configured */
		ast_log(LOG_WARNING, "no global section in %s\n", ASTDATA_CONFIG_FILE);
		ast_destroy(cfg);
		return(0);
	}

	while (var) {
		if (!strcasecmp(var->name, "sipfriends")) {
			ast_data_make_reservations(ASTDATA_QTYPE_SIP_FIND_PEER, var->value);
			ast_data_make_reservations(ASTDATA_QTYPE_SIP_FIND_PEERS_WITH_MAILBOX, var->value);
			ast_data_make_reservations(ASTDATA_QTYPE_SIP_FIND_USER, var->value);
			ast_data_make_reservations(ASTDATA_QTYPE_SIP_UPDATE_PEER, var->value);
		}
		else if (!strcasecmp(var->name, "iaxfriends")) {
			ast_data_make_reservations(ASTDATA_QTYPE_IAX_FIND_PEER, var->value);
			ast_data_make_reservations(ASTDATA_QTYPE_IAX_FIND_USER, var->value);
			ast_data_make_reservations(ASTDATA_QTYPE_IAX_UPDATE_PEER, var->value);
		}		
		else if (!strcasecmp(var->name, "extensionlookup")) {
			ast_data_make_reservations(ASTDATA_QTYPE_EXTENSION_LOOKUP, var->value);
		}
		else if (!strcasecmp(var->name, "voicemail")) {
			ast_data_make_reservations(ASTDATA_QTYPE_VM_FIND_USER, var->value);
			ast_data_make_reservations(ASTDATA_QTYPE_VM_CHANGE_PASSWORD,
				var->value);
		}
		else if (!strcasecmp(var->name, "directory")) {
			ast_data_make_reservations(ASTDATA_QTYPE_VM_DIRECTORY, var->value);
		}
		else if (!strcasecmp(var->name, "endofcall")) {
			ast_data_make_reservations(ASTDATA_QTYPE_END_OF_CALL, var->value);
		}
		var = var->next;
	}

	ast_destroy(cfg);
		
	return(0);
}

/***************************************************************************/

int
ast_data_unload(void) {
	struct ast_data_handleritem *hitem = NULL;

	ast_cli_unregister(&ast_data_show_handlers_cli);
	ast_cli_unregister(&ast_data_no_debug_cli);
	ast_cli_unregister(&ast_data_debug_cli);
	
	ast_mutex_lock(&ast_data_lock);

	/* Free everything... we are about to die */
	hitem = handlers;
	while (hitem) {
		handlers = hitem->next;
		free(hitem);
		hitem = handlers;
	}
	ast_mutex_unlock(&ast_data_lock);

	return(0);
}

/***************************************************************************/

int
ast_data_isqtype_registered(ASTDATA_QTYPE qtype) {
	struct ast_data_handleritem *hitem = NULL;

	hitem = handlers;
	while (hitem) {
		if ((hitem->qtype == qtype) && (hitem->query_handler != NULL)) {
			break;
		}
		hitem = hitem->next;
	}
	if (hitem) {
		 return(-1);
	}
	else {
		return(0);
	}
}

/***************************************************************************/

int
ast_data_isvalid_qtype(ASTDATA_QTYPE qtype) {
	int result = 0;

	switch (qtype) {
		case ASTDATA_QTYPE_EXTENSION_LOOKUP:
		case ASTDATA_QTYPE_SIP_FIND_USER:
		case ASTDATA_QTYPE_SIP_FIND_PEER:
		case ASTDATA_QTYPE_SIP_FIND_PEERS_WITH_MAILBOX:
		case ASTDATA_QTYPE_SIP_UPDATE_PEER:
		case ASTDATA_QTYPE_VM_CHANGE_PASSWORD:
		case ASTDATA_QTYPE_VM_FIND_USER:
		case ASTDATA_QTYPE_VM_DIRECTORY:
		case ASTDATA_QTYPE_IAX_FIND_USER:
		case ASTDATA_QTYPE_IAX_FIND_PEER:
		case ASTDATA_QTYPE_IAX_UPDATE_PEER:
		case ASTDATA_QTYPE_END_OF_CALL:
			result = -1;
			break;
	}
	return(result);
}

/***************************************************************************/

char *
ast_data_qtype2str(ASTDATA_QTYPE qtype) {
	switch (qtype) {
		/* Query type translations
		 * One case entry for each ASTDATA_xxxx
		 */
		case ASTDATA_QTYPE_EXTENSION_LOOKUP:
			return "EXTENSION_LOOKUP";
			break;
		case ASTDATA_QTYPE_SIP_FIND_USER:
			return "SIP_FIND_USER";
			break;
		case ASTDATA_QTYPE_SIP_FIND_PEER:
			return "SIP_FIND_PEER";
			break;
		case ASTDATA_QTYPE_SIP_FIND_PEERS_WITH_MAILBOX:
                        return "SIP_FIND_PEERS_WITH_MAILBOX";
                        break;
		case ASTDATA_QTYPE_SIP_UPDATE_PEER:
			return "SIP_UPDATE_PEER";
			break;
		case ASTDATA_QTYPE_VM_CHANGE_PASSWORD:
			return "VM_CHANGE_PASSWORD";
			break;
		case ASTDATA_QTYPE_VM_FIND_USER:
			return "VM_FIND_USER";
			break;
		case ASTDATA_QTYPE_VM_DIRECTORY:
			return "VM_DIRECTORY";
			break;
		case ASTDATA_QTYPE_IAX_FIND_USER:
			return "IAX_FIND_USER";
			break;
		case ASTDATA_QTYPE_IAX_FIND_PEER:
			return "IAX_FIND_PEER";
			break;
		case ASTDATA_QTYPE_IAX_UPDATE_PEER:
			return "IAX_UPDATE_PEER";
			break;
		case ASTDATA_QTYPE_END_OF_CALL:
			return "END_OF_CALL";
			break;
		default:
			return "UNDEFINED";
			break;
	}
}

/***************************************************************************/

ASTDATA_QINFO *
ast_data_find_qinfo(ASTDATA_DBINFO *dbinfo, ASTDATA_QTYPE qtype) {
	ASTDATA_QINFO *qinfo = NULL;
	int found = 0;

	if (dbinfo) {
		qinfo = dbinfo->qinfo;
		while (qinfo && !found) {
			if (qinfo->qtype == qtype) {
				found = 1;
			}
			else {
				qinfo = qinfo->next;
			}
		}
	}
	return(qinfo);
}

/***************************************************************************/

char *
ast_data_set_qinfo_query(ASTDATA_DBINFO *dbinfo, ASTDATA_QTYPE qtype, char *query, int usealtquery) {
	ASTDATA_QINFO *qinfo = NULL;

	qinfo = ast_data_find_qinfo(dbinfo, qtype);
	if (!qinfo) {
		/* qtype not present in this dbinfo */
		return(NULL);
	}
	else {
		if (usealtquery) {
			if (qinfo->altquery) {
				/* previous query stored, free it up to
				 * make room for its replacement
				 */
				free(qinfo->altquery);
				qinfo->altquery = NULL;
			}
			if (query) {
				qinfo->altquery = strdup(query);
			}
			return(qinfo->altquery);
		}
		else {
			if (qinfo->query) {
				/* previous query stored, free it up to
				 * make room for its replacement
				 */
				free(qinfo->query);
				qinfo->query = NULL;
			}
			if (query) {
				qinfo->query = strdup(query);
			}
			return(qinfo->query);
		}
	}
}

/***************************************************************************/

ASTDATA_QINFO *
ast_data_add_qinfo(ASTDATA_DBINFO *dbinfo, ASTDATA_QTYPE qtype, char *query) {
	ASTDATA_QINFO *qi = NULL;

	if (dbinfo) {
		qi = calloc(1, sizeof(ASTDATA_QINFO));
		if (!qi) {
			ast_log(LOG_ERROR, "Out of memory\n");
		}
		else {
			qi->qtype = qtype;
			if (query) {
				/* query was supplied, so allocate and copy it as well */
				qi->query = strdup(query);
				if (!qi->query) {
					ast_log(LOG_ERROR, "Out of memory\n");
					free(qi);
					qi = NULL;
				}
			}
			if (qi) {
				/* allocation went OK, now add to list */
				qi->next = dbinfo->qinfo;
				dbinfo->qinfo = qi;
			}
		}
	}
	return(qi);
}

/***************************************************************************/

ASTDATA_QINFO *
ast_data_add_qinfo_conditional(ASTDATA_DBINFO *dest, ASTDATA_DBINFO *check, ASTDATA_QTYPE qtype, char *query) {
	if (ast_data_find_qinfo(check, qtype)) {
		return(ast_data_add_qinfo(dest, qtype, query));
	}
	else {
		return(NULL);
	}
}

/***************************************************************************/

ASTDATA_ERROR
ast_data_parse_cfg_qtype(char *qtypename, ASTDATA_DBINFO *dest, ASTDATA_DBINFO *check) {
	if (!strcasecmp(qtypename, "sipfriends")) {
		/* Allow all SIP style queries to this db, if the
		 * allowed queries includes SIP
		 */
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_SIP_FIND_PEER, NULL);
		ast_data_add_qinfo_conditional(dest, check,
                        ASTDATA_QTYPE_SIP_FIND_PEERS_WITH_MAILBOX, NULL);
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_SIP_FIND_USER, NULL);
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_SIP_UPDATE_PEER, NULL);
	}
	else if (!strcasecmp(qtypename, "voicemail")) {
		/* Allow all Voicemail style queries to this db, if the
		 * allowed queries includes Voicemail
		 */
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_VM_FIND_USER, NULL);
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_VM_CHANGE_PASSWORD, NULL);
	}
	else if (!strcasecmp(qtypename, "iaxfriends")) {
		/* Allow all IAX style queries to this db, if the
		 * allowed queries includes IAX
		 */
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_IAX_FIND_PEER, NULL);
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_IAX_FIND_USER, NULL);
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_IAX_UPDATE_PEER, NULL);
	}
	else if (!strcasecmp(qtypename, "directory")) {
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_VM_DIRECTORY, NULL);
	}
	else if (!strcasecmp(qtypename, "endofcall")) {
		ast_data_add_qinfo_conditional(dest, check,
			ASTDATA_QTYPE_END_OF_CALL, NULL);
	}
	else {
		/* Ignore "default", and don't print a warning either */
		if (strcasecmp(qtypename, "default")) {
			ast_log(LOG_WARNING, "Unknown QType=%s in %s--Ignored.\n", qtypename,
				ASTDATA_CONFIG_FILE);
			return(ASTDATA_E_OTHER);
		}
	}
	return(ASTDATA_E_OK);
}

/***************************************************************************/

ASTDATA_ERROR
ast_data_free_qinfo(ASTDATA_DBINFO *dbinfo) {
	ASTDATA_QINFO	*next_qinfo = NULL;

	if (dbinfo) {
		while (dbinfo->qinfo) {
			next_qinfo = dbinfo->qinfo->next;
			if (dbinfo->qinfo->query) {
				/* free the dynamically allocated space for the query */
				free(dbinfo->qinfo->query);

				dbinfo->qinfo->query = NULL;
			}
			free(dbinfo->qinfo);
			dbinfo->qinfo = next_qinfo;
		}
	}
	return(ASTDATA_E_OK);
}

/***************************************************************************/

ASTDATA_ERROR
ast_data_make_reservations(ASTDATA_QTYPE qtype, char *qorder) {
	char *data = NULL;
	char *p = NULL;

	/* we don't want to destroy the incoming string, so lets make our
	 * own copy to play with
	 */
	data = ast_strdupa(qorder);
	if (!data) {
		/* no memory */
		ast_log(LOG_ERROR, "no memory available!\n");
		return(ASTDATA_E_NOMEM);
	}

	/* tokenize and allow space, comma, tab, and pipe as delimiters */
	p = strtok(data, " ,\t|");
	while (p) {
		if (ast_data_reserve(p, qtype) == ASTDATA_E_OK) {
			p = strtok(NULL, " ,\t|");
		}
		else {
			p = NULL;
		}
	}
	return(ASTDATA_E_OK);
}

/***************************************************************************/

ASTDATA_ERROR
ast_data_reserve(char *name, ASTDATA_QTYPE qtype) {
	struct ast_data_handleritem *hitem = NULL;
	struct ast_data_handleritem *i = NULL;

	/* Validate parameters */
	if (!name) {
		return(ASTDATA_E_BAD_PARAM);
	}

	if (!ast_data_isvalid_qtype(qtype)) {
		ast_log(LOG_WARNING, "Data Handler %s, Query Type %d Undefined\n",
				name, qtype);
		return(ASTDATA_E_BAD_PARAM);
	}

	/* Find if Handler reservation already exists */
	ast_mutex_lock(&ast_data_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == qtype && (!strcmp(hitem->name, name))) {
			break;
		}
		hitem = hitem->next;
	}

	ast_mutex_unlock(&ast_data_lock);

	if (hitem) {
		/* reservation already exists  */
		ast_log(LOG_WARNING,
			"Already have Query Type %s reserved for handler %s.\n",
			ast_data_qtype2str(qtype), hitem->name);
		return(ASTDATA_E_OTHER);
	}

	/* Allocate memory and setup struct */
	hitem = calloc(1, sizeof(struct ast_data_handleritem));
	if (!hitem) {
		/* out of memory */
		return(ASTDATA_E_NOMEM);
	}

	strncpy(hitem->name, name, sizeof(hitem->name) - 1);
	hitem->qtype = qtype;

	/* Add to chain (at end on purpose to preserve registration order) */
	ast_mutex_lock(&ast_data_lock);
	if (handlers == NULL) {
		/* empty list so far, add at front */
		handlers = hitem;
	}
	else {
		/* move to end of list, and add it there */
		for (i = handlers; i->next; i = i->next)
			;
		i->next = hitem;
	}
	ast_mutex_unlock(&ast_data_lock);

	ast_verbose(VERBOSE_PREFIX_2
		"Data: Reservation made for \"%s\" query type %s\n",
		name, ast_data_qtype2str(qtype));
	return(ASTDATA_E_OK);
}

/***************************************************************************/

ASTDATA_ERROR
ast_data_register(char *cliname, char *techname, ASTDATA_DBINFO *dbinfo, ASTDATA_QTYPE qtype, ast_data_query_handler handler)
{
	char desc[200];
	char *handlername = NULL;

	struct ast_data_handleritem *hitem = NULL;

	/* Validate parameters */
	if (!cliname || !techname || !dbinfo) {
		return(ASTDATA_E_BAD_PARAM);
	}

	if (!ast_data_isvalid_qtype(qtype)) {
		ast_log(LOG_WARNING, "Data Handler %s, Query Type %d Undefined\n",
			cliname, qtype);
		return(ASTDATA_E_BAD_PARAM);
	}

	switch (qtype) {
		case ASTDATA_QTYPE_SIP_FIND_PEER:
			handlername = "SIP Peer Lookup";
			break;
		case ASTDATA_QTYPE_SIP_FIND_PEERS_WITH_MAILBOX:
                        handlername = "SIP Peers With Mailbox Lookup";
                        break;
		case ASTDATA_QTYPE_SIP_FIND_USER:
			handlername = "SIP User Lookup";
			break;
		case ASTDATA_QTYPE_SIP_UPDATE_PEER:
			handlername = "SIP Peer Update";
			break;
		case ASTDATA_QTYPE_IAX_FIND_PEER:
			handlername = "IAX Peer Lookup";
			break;
		case ASTDATA_QTYPE_IAX_FIND_USER:
			handlername = "IAX User Lookup";
			break;
		case ASTDATA_QTYPE_IAX_UPDATE_PEER:
			handlername = "IAX Update Peer";
			break;
		case ASTDATA_QTYPE_VM_FIND_USER:
			handlername = "Voicemail User Lookup";
			break;
		case ASTDATA_QTYPE_VM_CHANGE_PASSWORD:
			handlername = "Voicemail Password Change";
			break;
		case ASTDATA_QTYPE_VM_DIRECTORY:
			handlername = "Voicemail Directory";
			break;
		case ASTDATA_QTYPE_EXTENSION_LOOKUP:
			handlername = "Extension Lookup";
			break;
		case ASTDATA_QTYPE_END_OF_CALL:
			handlername = "End of Call";
			break;
	}

	if (!ast_data_find_qinfo(dbinfo, qtype)) {
		return(ASTDATA_E_NOT_FOUND);
	}

	snprintf(desc, sizeof(desc), "%s %s Handler", techname, handlername);
		
	if (!handler) {
		ast_log(LOG_WARNING, "Data Handler %s, Query Type %s lacks handler\n",
			cliname, ast_data_qtype2str(qtype));
		return(ASTDATA_E_BAD_PARAM);
	}

	/* Find Handler reservation */
	ast_mutex_lock(&ast_data_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == qtype && (!strcmp(hitem->name, cliname))) {
			break;
		}
		hitem = hitem->next;
	}

	ast_mutex_unlock(&ast_data_lock);

	if (!hitem) {
		/* no reservation found */
		ast_log(LOG_WARNING,
			"No reservation found for Query Type %s under handler %s.\n",
			ast_data_qtype2str(qtype), cliname);
		return(ASTDATA_E_OTHER);
	}

	if (hitem->query_handler) {
		ast_log(LOG_WARNING,
			"Already have Query Type %s registered to handler %s.\n",
			ast_data_qtype2str(qtype), cliname);
		return(ASTDATA_E_OTHER);
	}

	/* make the assignment for the handler */
	strncpy(hitem->desc, desc, sizeof(hitem->desc) - 1);
	hitem->query_handler = handler;

	ast_verbose(VERBOSE_PREFIX_2
		"\"%s\" registration completed for query type %s\n",
		cliname, ast_data_qtype2str(qtype));

	return(ASTDATA_E_OK);
}

/***************************************************************************/

void
ast_data_unregister(char *name, ASTDATA_QTYPE qtype)
{
	struct ast_data_handleritem *hitem = NULL;

	if (!name) {
		ast_log(LOG_WARNING,
			"Missing name attempting to unregister a Data Resource Handler.\n");
		return;
	}

	if (!ast_data_isvalid_qtype(qtype)) {
		ast_log(LOG_WARNING,
			"Attempt to unregister invalid query type %d from %s Data Resource Handler\n",
				qtype, name);
		return;
	}

	
	ast_mutex_lock(&ast_data_lock);
	hitem = handlers;
	while (hitem) {
		if (!strcasecmp(name, hitem->name) && (hitem->qtype == qtype)) {
			if (hitem->query_handler) {
				ast_verbose(VERBOSE_PREFIX_2
					"Data Resource: %s unregistered query type %s\n",
					name, ast_data_qtype2str(hitem->qtype));
				hitem->query_handler = NULL;
			}
			break;
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&ast_data_lock);
}

/***************************************************************************/

void
ast_data_unregister_all(char *name) {
		ast_verbose(VERBOSE_PREFIX_2
			"Unregistering all known %s Data Resource Handlers\n",
			name);
		ast_data_unregister(name, ASTDATA_QTYPE_EXTENSION_LOOKUP);
		ast_data_unregister(name, ASTDATA_QTYPE_SIP_FIND_USER);
		ast_data_unregister(name, ASTDATA_QTYPE_SIP_FIND_PEER);
		ast_data_unregister(name, ASTDATA_QTYPE_SIP_FIND_PEERS_WITH_MAILBOX);
		ast_data_unregister(name, ASTDATA_QTYPE_SIP_UPDATE_PEER);
		ast_data_unregister(name, ASTDATA_QTYPE_VM_CHANGE_PASSWORD);
		ast_data_unregister(name, ASTDATA_QTYPE_VM_FIND_USER);
		ast_data_unregister(name, ASTDATA_QTYPE_VM_DIRECTORY);
		ast_data_unregister(name, ASTDATA_QTYPE_IAX_FIND_USER);
		ast_data_unregister(name, ASTDATA_QTYPE_IAX_FIND_PEER);
		ast_data_unregister(name, ASTDATA_QTYPE_IAX_UPDATE_PEER);
		ast_data_unregister(name, ASTDATA_QTYPE_END_OF_CALL);
}

/***************************************************************************/

ASTDATA_PACKET *
ast_data_do_query(ASTDATA_PACKET *request)
{
	struct ast_data_handleritem *hitem;
	ASTDATA_PACKET *response = NULL;

	if (!request) {
		return(response);
	}

	if (astdatadebug) {
		ast_verbose(VERBOSE_PREFIX_2 "Request:\n");
		ast_data_dump_packet(request);	
	}

	ast_mutex_lock(&ast_data_lock);
	hitem = handlers;
	/* step through the handlers looking for ones
	 * that handle the same query type as the handler wants
	 */
	while (hitem) {
		if (hitem->qtype == request->qtype) {
			/* send the request packet to the handler.
			 * handler will return NULL if not found/no success
			 */
			if (hitem->query_handler) {
				response = hitem->query_handler(request);
				if (response) {
					/* the handler had a response.  stop if
					 * the errormsg is OK
					 */
					if (ast_data_error_number(response) == ASTDATA_E_OK) {
						break;
					}
					else {
						ast_data_free_packet(response);
						response = NULL;
					}
				}
			}
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&ast_data_lock);

	if (astdatadebug) {
		ast_verbose(VERBOSE_PREFIX_2 "Response:\n");
		ast_data_dump_packet(response);
	}
	
	return(response);
}

/***************************************************************************/

void
ast_data_set_error(ASTDATA_PACKET *packet, ASTDATA_ERROR errnum, char *errmsg) {
	ASTDATA_PAIR *vp = NULL;
   if (!packet) return;

	/* find if error number already in packet, and replace or add as needed */
	vp = ast_data_first_pair(packet, ASTDATA_ATTR_ERR_NUM);
	if (vp == NULL) {
		ast_data_add_integer(packet, ASTDATA_ATTR_ERR_NUM, errnum);
	}
	else {
		vp->lvalue = errnum;
	}

	/* do same with the error message */
	vp = ast_data_first_pair(packet, ASTDATA_ATTR_ERR_MSG);
	if (vp == NULL) {
		ast_data_add_string(packet, ASTDATA_ATTR_ERR_MSG, errmsg);
	}
	else {
		if (vp->strvalue) {
			free(vp->strvalue);
		}
		if (errmsg) {
			vp->strvalue = strdup(errmsg);
		}
		else {
			vp->strvalue = strdup("missing error message");
		}
	}
}

/***************************************************************************/

void
ast_data_clear_error(ASTDATA_PACKET *packet) {
	ast_data_set_error(packet, ASTDATA_E_OK, "OK");
}

/***************************************************************************/

void
ast_data_free_packet(ASTDATA_PACKET *packet) {
	ASTDATA_PAIR *next;

	if (!packet) {
		return;
	}

	ast_mutex_lock(&packet->lock);
	while (packet->data != NULL) {
		next = packet->data->next;
		if (packet->data->strvalue) {
			free(packet->data->strvalue);
		}
		free(packet->data);
		packet->data = next;
	}
	
	ast_mutex_unlock(&packet->lock);
	free(packet);
}

/***************************************************************************/

ASTDATA_PAIR *
ast_data_first_pair(ASTDATA_PACKET *packet, ASTDATA_ATTR attribute) {
	ASTDATA_PAIR *item;

	if (!packet) {
		return(NULL);
	}

	ast_mutex_lock(&packet->lock);
	item = packet->data;
	while (item && item->attribute != attribute) {
		item = item->next;
	}
	ast_mutex_unlock(&packet->lock);

	return(item);
}

/***************************************************************************/

ASTDATA_PAIR *
ast_data_next_pair(ASTDATA_PAIR *prevpair, ASTDATA_ATTR attribute) {
	ASTDATA_PAIR *item;

	if (!prevpair) {
		return(NULL);
	}

	item = prevpair->next;
	while (item && item->attribute != attribute) {
		item = item->next;
	}

	return(item);
}

/***************************************************************************/

void
ast_data_delete_pair(ASTDATA_PACKET *packet, ASTDATA_ATTR attribute) {
	ASTDATA_PAIR *i, *next, *last = NULL;

	if (!packet) {
		return;
	}

	ast_mutex_lock(&packet->lock);
	for (i = packet->data; i; i = next) {
		next = i->next;
		if (i->attribute == attribute) {
			if (last) {
				last->next = next;
			}
			else {
				packet->data = next;
			}
			if (i->strvalue) {
				free(i->strvalue);
			}
			free(i);
		}
		else {
			last = i;
		}
	}
	ast_mutex_unlock(&packet->lock);
}

/***************************************************************************/

/* Adds a ASTDATA_PAIR at the end of a chain in a packet
 * --Added at end on purpose in order to preserve order 
 * --Would have been too easy to add at front
 */
void
ast_data_add_pair(ASTDATA_PACKET *packet, ASTDATA_PAIR *new) {
	ASTDATA_PAIR *i;

	if (!packet) {
		return;
	}

	ast_mutex_lock(&packet->lock);
	new->next = NULL;
	if (packet->data == NULL) {
		packet->data = new;
		ast_mutex_unlock(&packet->lock);
		return;
	}
	for (i = packet->data; i->next; i = i->next)
		;
	i->next = new;
	ast_mutex_unlock(&packet->lock);
}

/***************************************************************************/

ASTDATA_PAIR *
ast_data_add_integer(ASTDATA_PACKET *packet, ASTDATA_ATTR attr, int lvalue) {
	ASTDATA_PAIR *vp = NULL;
	ASTDATA_TYPE datatype = ASTDATA_TYPE_INVALID;

	if (!packet) {
		/* packet is not a pointer */
		return(vp);
	}

	datatype = ast_data_attr_type(attr);
	if ((datatype != ASTDATA_TYPE_INTEGER) && (datatype != ASTDATA_TYPE_DATE)) {
		/* attribute is not an integer or a date */
		return(vp);
	}

	vp = calloc(1, sizeof(ASTDATA_PAIR));
	if (vp) {
		vp->attribute = attr;
		vp->lvalue = lvalue;

		ast_data_add_pair(packet, vp);
	}

	return(vp);
}

/***************************************************************************/

ASTDATA_PAIR *
ast_data_add_date(ASTDATA_PACKET *packet, ASTDATA_ATTR attr, int datevalue) {
	/* simple wrapper for ast_data_add_integer */
	return(ast_data_add_integer(packet, attr, datevalue));
}

/***************************************************************************/

ASTDATA_PAIR *
ast_data_add_string(ASTDATA_PACKET *packet, ASTDATA_ATTR attr, char *strvalue) {
	ASTDATA_PAIR *vp = NULL;
	ASTDATA_TYPE datatype = ASTDATA_TYPE_INVALID;

	if (!packet) {
		/* no packet */
		return(vp);
	}

	datatype = ast_data_attr_type(attr);
	if (datatype != ASTDATA_TYPE_STRING) {
		/* attribute is not a string */
		return(vp);
	}

	if (strvalue == NULL) {
		/* null ptr input */
		return(vp);
	}

	vp = calloc(1, sizeof(ASTDATA_PAIR));
	if (vp) {
		vp->attribute = attr;
		vp->strvalue = strdup(strvalue);

		ast_data_add_pair(packet, vp);
	}

	return(vp);
}

/***************************************************************************/

ASTDATA_PAIR *
ast_data_add_ipaddr(ASTDATA_PACKET *packet, ASTDATA_ATTR attr, char *ipaddr) {
	ASTDATA_PAIR *vp = NULL;
	ASTDATA_TYPE datatype = ASTDATA_TYPE_INVALID;

	if (!packet) {
		/* no packet */
		return(vp);
	}

	datatype = ast_data_attr_type(attr);
	if (datatype != ASTDATA_TYPE_IPADDR) {
		/* attribute is not an IP Address */
		return(vp);
	}

	vp = calloc(1, sizeof(ASTDATA_PAIR));
	if (vp) {
		vp->attribute = attr;
		inet_aton(ipaddr, &vp->ipaddr);

		ast_data_add_pair(packet, vp);
	}

	return(vp);
}

/***************************************************************************/

ASTDATA_ATTR
ast_data_check_params(ASTDATA_PACKET *packet, ...) {
	ASTDATA_PAIR *vp = NULL;
	ASTDATA_ATTR attribute = ASTDATA_ATTR_INVALID;
	ASTDATA_ERROR result = ASTDATA_E_OK;
	ASTDATA_TYPE datatype = ASTDATA_TYPE_INVALID;
	va_list ap;
	int found = 0;
	int foundcount = 0;
	int attrcount = 0;

	if (packet == NULL) {
		return(ASTDATA_E_BAD_PARAM);
	}

	va_start(ap, packet);
	attribute = va_arg(ap, ASTDATA_ATTR);
	
	while (attribute != ASTDATA_ATTR_INVALID) {
		datatype = ast_data_attr_type(attribute);
		attrcount++;
		found = 0;
		/*check all pairs to see if at least one matches "attribute" */
		vp = ast_data_first_pair(packet, attribute);

		if (vp) {
			switch (datatype) {
				case ASTDATA_TYPE_STRING:
					/* for strings, we make sure the strvalue ptr is good as well */
					if (vp->strvalue) {
						found = 1;
					}
					break;
				default:
					found = 1;
					break;
			} /* end switch(datatype) */
		} /* end if (vp) */

		if (found) {
			foundcount++;
			attribute = va_arg(ap, ASTDATA_ATTR);
		}
		else {
#if 0
			ast_log(LOG_WARNING, "missing attribute '%s' in %s query.\n",
				ast_data_attr_name(attribute),
				ast_data_qtype2str(packet->qtype));
#endif
			/* missing one attribute at least, short circuit the while
			 * loop for speed
			 */
			result = attribute;
			attribute = ASTDATA_ATTR_INVALID;
		}
	} /* end while(attribute != invalid) */
	va_end(ap);

	return(result);
}

/***************************************************************************/

ASTDATA_ATTR
ast_data_extract_packet(ASTDATA_PACKET *packet, ...) {
	ASTDATA_PAIR *vp = NULL;
	ASTDATA_ATTR attribute = ASTDATA_ATTR_INVALID;
	ASTDATA_ERROR result = ASTDATA_E_OK;
	ASTDATA_TYPE datatype = ASTDATA_TYPE_INVALID;
	va_list ap;
	int foundcount = 0;
	int attrcount = 0;

	int *intptr = NULL;
	char **charptr = NULL;
	struct in_addr *inaddrptr = NULL;

	if (packet == NULL) {
		return(ASTDATA_E_BAD_PARAM);
	}

	va_start(ap, packet);
	attribute = va_arg(ap, ASTDATA_ATTR);
	
	while (attribute != ASTDATA_ATTR_INVALID) {
		datatype = ast_data_attr_type(attribute);
		attrcount++;
		/* check all pairs to see if at least one matches "attribute" */
		vp = ast_data_first_pair(packet, attribute);

		switch (datatype) {
			case ASTDATA_TYPE_STRING:
				/* for strings, we make sure the strvalue ptr is good as well */
				charptr = va_arg(ap, char **);
				if (vp) {
					if (vp->strvalue) {
						*charptr = vp->strvalue;
					}
					else {
						*charptr = NULL;
					}
				}
				break;

			case ASTDATA_TYPE_INTEGER:
			case ASTDATA_TYPE_DATE:
				intptr = va_arg(ap, int *);
				if (vp) {
					*intptr = vp->lvalue;
				}
				break;

			case ASTDATA_TYPE_IPADDR:
				inaddrptr = va_arg(ap, struct in_addr *);
				if (vp) {
					*inaddrptr = vp->ipaddr;
				}
				break;

		} /* end switch(datatype) */

		if (vp) {
			foundcount++;
		}
		else {
#if 0
			ast_log(LOG_WARNING, "missing attribute '%s' in %s query.\n",
				ast_data_attr_name(attribute),
				ast_data_qtype2str(packet->qtype));
#endif
			/* Mark the attribute as missing for return to calling function */
			result = attribute;
		}

		attribute = va_arg(ap, ASTDATA_ATTR);
	} /* end while(attribute != invalid) */
	va_end(ap);

	return(result);
}


/***************************************************************************/

ASTDATA_PACKET *
ast_data_create_packet(ASTDATA_QTYPE qtype, ...) {
	ASTDATA_PACKET *result = NULL;
	ASTDATA_ATTR attribute = ASTDATA_ATTR_INVALID;
	ASTDATA_TYPE datatype = ASTDATA_TYPE_INVALID;
	ASTDATA_PAIR *vp = NULL;
	va_list ap;

	if (!ast_data_isvalid_qtype(qtype)) {
		/* invalid query type */
		ast_log(LOG_ERROR, "invalid qtype %d\n", qtype);
		return(NULL);
	}

	/* create the packet */
	result = calloc(1, sizeof(ASTDATA_PACKET));
	if (result == NULL) {
		/* out of memory */
		ast_log(LOG_ERROR, "out of memory!\n");
		return(NULL);
	}
	else {
		/* initialize the packet */
		result->qtype = qtype;

		va_start(ap, qtype);
		do {
			/*  args are coming in as ATTRIBUTE, VALUE, ATTRIBUTE, VALUE, ... */
			attribute = va_arg(ap, ASTDATA_ATTR);
			datatype = ast_data_attr_type(attribute);

			/* add the value to the packet the proper way */
			switch (datatype) {
				case ASTDATA_TYPE_DATE:
				case ASTDATA_TYPE_INTEGER:
					vp = ast_data_add_integer(result, attribute, va_arg(ap, int));
					break;
				case ASTDATA_TYPE_STRING:
					vp = ast_data_add_string(result, attribute, va_arg(ap, char *));
					break;
				case ASTDATA_TYPE_IPADDR:
					vp = ast_data_add_ipaddr(result, attribute, va_arg(ap, char *));
					break;
				default:
					/* invalid datatype / end of list */
					attribute = ASTDATA_ATTR_INVALID;
					break;
			} /* end switch */

			if (vp == NULL) {
				/* pair could not be added successfully... clear memory return NULL
				 */
				ast_log(LOG_ERROR, "could not add value for attribute %d(%s)\n",
					attribute, ast_data_attr_name(attribute));
				ast_data_free_packet(result);
				attribute = ASTDATA_ATTR_INVALID;
				result = NULL;
			}
		} while (attribute != ASTDATA_ATTR_INVALID);
		va_end(ap);
		return(result);
	}
}

/***************************************************************************/

char *
ast_data_error_message(ASTDATA_PACKET *packet) {
	ASTDATA_PAIR *vp = NULL;

	vp = ast_data_first_pair(packet, ASTDATA_ATTR_ERR_MSG);
	if (vp) {
		return(vp->strvalue);
	}
	else {
		return(NULL);
	}
}

/***************************************************************************/

ASTDATA_ERROR
ast_data_error_number(ASTDATA_PACKET *packet) {
	ASTDATA_PAIR *vp = NULL;

	vp = ast_data_first_pair(packet, ASTDATA_ATTR_ERR_NUM);
	if (vp) {
		return(vp->lvalue);
	}
	else {
		return(ASTDATA_E_NOT_PRESENT);
	}
}

/***************************************************************************/

void
ast_data_dump_packet(ASTDATA_PACKET *packet) {
	ASTDATA_PAIR *item;
	char iabuf[INET_ADDRSTRLEN];

	/* TODO: jump out if debug or verbose level not high enough */

	if (!packet) {
		return;
	}

	ast_mutex_lock(&packet->lock);
	ast_verbose(VERBOSE_PREFIX_2 "==== Packet Dump ====\n");
	ast_verbose(VERBOSE_PREFIX_2 "Query Type: %s\n",
		ast_data_qtype2str(packet->qtype));

	item = packet->data;
	while (item) {
		/* change the output method based on the attribute's data type */
		switch (ast_data_attr_type(item->attribute)) {
			case ASTDATA_TYPE_STRING:
				ast_verbose(VERBOSE_PREFIX_2
					"(%d)%s = (STR)\"%s\"\n", item->attribute,
					ast_data_attr_name(item->attribute),
					item->strvalue);
				break;

			case ASTDATA_TYPE_INTEGER:
				ast_verbose(VERBOSE_PREFIX_2
					"(%d)%s = (INT)%d\n", item->attribute,
					ast_data_attr_name(item->attribute),
					item->lvalue);
				break;

			case ASTDATA_TYPE_IPADDR:
				ast_verbose(VERBOSE_PREFIX_2
					"(%d)%s = (IP)%s\n", item->attribute,
					ast_data_attr_name(item->attribute),
					ast_inet_ntoa(iabuf, sizeof(iabuf), item->ipaddr));
				break;

			/* TODO: change output so lvalue comes out as a date */
			case ASTDATA_TYPE_DATE:
				ast_verbose(VERBOSE_PREFIX_2
					"(%d)%s = (DATE)%d\n", item->attribute,
					ast_data_attr_name(item->attribute),
					item->lvalue);
				break;

			default:
				ast_verbose(VERBOSE_PREFIX_2
					"(%d)%s = (UNDEFINED TYPE)(INT?)%d (IP?)%s (STR?)\"%s\"\n",
					item->attribute,
					ast_data_attr_name(item->attribute),
					item->lvalue, 
					ast_inet_ntoa(iabuf, sizeof(iabuf), item->ipaddr),
					item->strvalue);
				break;
		} /* end switch */
		item = item->next;
	}
	ast_verbose(VERBOSE_PREFIX_2 "==== End of Packet Dump ====\n");
	ast_mutex_unlock(&packet->lock);
}

/***************************************************************************/

char *
ast_data_attr_name(ASTDATA_ATTR attribute) {
	int i;
	char *result = "undefined";

	for (i = 0; attribute_dictionary[i].type != ASTDATA_TYPE_INVALID; i++) {
		if (attribute_dictionary[i].attribute == attribute) {
			result = attribute_dictionary[i].name;
		}
	}
	return(result);
}

/***************************************************************************/

ASTDATA_ATTR
ast_data_name2attr(char *name) {
	int i;
	ASTDATA_ATTR result = ASTDATA_ATTR_INVALID;

	for (i = 0; attribute_dictionary[i].type != ASTDATA_TYPE_INVALID; i++) {
		if (!strcasecmp(attribute_dictionary[i].name, name)) {
			result = attribute_dictionary[i].attribute;
		}
	}
	return(result);
}

/***************************************************************************/

ASTDATA_TYPE
ast_data_attr_type(ASTDATA_ATTR attribute) {
	int i;
	ASTDATA_TYPE result = ASTDATA_TYPE_INVALID;

	for (i = 0; attribute_dictionary[i].type != ASTDATA_TYPE_INVALID; i++) {
		if (attribute_dictionary[i].attribute == attribute) {
			result = attribute_dictionary[i].type;
		}
	}
	return(result);
}

/***************************************************************************/

int
ast_data_reload(void) {
	ast_data_unload();
	ast_data_load();

	return(0);
}

/***************************************************************************/

char *ast_data_digit2regexp(char digit) {
	switch (digit) {
		case '0':
			return("0");
			break;
		case '1':
			return("1");
			break;
		case '2':
			return("(a|b|c|A|B|C)");
			break;
		case '3':
			return("(d|e|f|D|E|F)");
			break;
		case '4':
			return("(g|h|i|G|H|I)");
			break;
		case '5':
			return("(j|k|l|J|K|L)");
			break;
		case '6':
			return("(m|n|o|M|N|O)");
			break;
		case '7':
			return("(p|q|r|s|P|Q|R|S)");
			break;
		case '8':
			return("(t|u|v|T|U|V)");
			break;
		case '9':
			return("(w|x|y|z|W|X|Y|Z)");
			break;
		case '*':
		case '#':
			return("sillywordnomatch");
			break;
		default:
			return("");
			break;
	}
}

/***************************************************************************/

int ast_data_dtmf2regexp(char *buff, int bufflen, char *dtmf) {
	int i = 0;
	char *fmtbuff = NULL;

	*buff = '\0';
	fmtbuff = alloca(bufflen);
	/* The starting snprintf pattern to use */
	strncpy(fmtbuff, "%s%%s", bufflen-1);

	/* for each character in 'dtmf', append to the buffer the
	 * portion of the regexp string needed to match letters to the
	 * digit
	 */
	for (i = 0; i<strlen(dtmf); i++) {
		snprintf(buff, bufflen, fmtbuff, ast_data_digit2regexp(dtmf[i]));
		snprintf(fmtbuff, bufflen, "%s%%%%s", buff);
	}

	/* now find the last %s in the string and eliminate it */
	if ((fmtbuff = strrchr(buff, '%'))) {
		*fmtbuff = '\0';
	}

	return(strlen(buff));
}

/***************************************************************************/

char *ast_data_strrepl(char *dest, int destsize, char *src, char *find, char *replace) {
	char *find_loc = NULL;
	char c;
	int bytes = 0;
	int max_copy = 0;
	char *dest_pos = NULL;
	char *src_pos = NULL;
	char *lcase_src = NULL;
	char *lcase_src_pos = NULL;
	char *lcase_find = NULL;

	max_copy = destsize;
	dest_pos = dest;
	src_pos = src;

	/* create lower case versions of src and find, for case-insensitive
	 * replacements
	 */
	lcase_src = strdup(src);
	lcase_find = strdup(find);
	if (!lcase_src || !lcase_find) {
		ast_log(LOG_ERROR, "Out of memory\n");
		if (lcase_src) free(lcase_src);
		if (lcase_find) free(lcase_find);
		return(NULL);
	}
	
	lcase_src_pos = lcase_src;
	for (bytes=0; bytes<strlen(lcase_src); bytes++)
		lcase_src[bytes] = tolower(lcase_src[bytes]);
	for (bytes=0; bytes<strlen(lcase_find); bytes++)
		lcase_find[bytes] = tolower(lcase_find[bytes]);

	memset(dest, 0, destsize);

	/* locate the 'find' string inside 'src' */
	find_loc = strstr(lcase_src_pos, lcase_find);
	while (find_loc) {

		/* save and null out character so snprintf works ok */
		c = src_pos[find_loc - lcase_src_pos];
		src_pos[find_loc - lcase_src_pos] = '\0';
		bytes = snprintf(dest_pos, max_copy, "%s%s", src_pos, replace);
	
		/* put the character back */
		// src_pos[find_loc - src_pos] = c;
		src_pos[find_loc - lcase_src_pos] = c;

		/* reduce the max bytes that can now be copied */
		max_copy -= bytes;
		dest_pos += bytes;

		/* set the place to start looking for the next occurance */
		bytes = find_loc - lcase_src_pos;
		src_pos += (bytes + strlen(find));
		lcase_src_pos += (bytes + strlen(lcase_find));

		/* find the next occurance */
		find_loc = strstr(lcase_src_pos, lcase_find);
	}

	/* copy the end of the string after all pattern matches */
	snprintf(dest_pos, max_copy, "%s", src_pos);
	
	free(lcase_src);
	free(lcase_find);

	return(dest);
}

/***************************************************************************/

char *ast_data_var_replace(char *dest, int destsize, char *template, ASTDATA_PACKET *pkt) {
	char *buff = NULL;
	int numbytes = 0;
	char *v = NULL;
	char *v_end = NULL;
	char pattern[80] = "";
	char v_name[80] = "";
	char strbuff[50] = "";
	ASTDATA_ATTR attr = ASTDATA_ATTR_INVALID;
	ASTDATA_PAIR *pair = NULL;

	/* make another buffer, the same size as dest */
	buff = calloc(1, destsize);
	if (!buff) {
		ast_log(LOG_ERROR, "Out of memory\n");
		return(NULL);
	}
	strncpy(dest, template, destsize);

	v = strstr(template, "${");
	while (v) {
		v_end = strchr(v, '}');
		if (!v_end) {
			/* no closing '}' to go with "${", so cannot complete job here
			 * set v NULL so while bails
			 */
			v = NULL;
		}
		else {
			numbytes = v_end - v + 1;
			if (numbytes > sizeof(pattern)) {
				numbytes = sizeof(pattern);
			}  
			memset(v_name, 0, sizeof(v_name));
			memcpy(v_name, v + 2, numbytes - 3);

			attr = ast_data_name2attr(v_name);
			if (attr == ASTDATA_ATTR_INVALID) {
				/* invalid name, so stop replacing  and bail out */
				v = NULL;
			}
			else {
				pair = ast_data_first_pair(pkt, attr);
				if (!pair) {
					/* attribute from name is ok, but no attribute like that
					 * in this packet
					 */
					v = NULL;
				}
				else {
					memset(pattern, 0, sizeof(pattern));
					memset(strbuff, 0, sizeof(strbuff));
					memcpy(pattern, v, numbytes);

					/* get the value from the pair, as a string */
					switch (ast_data_attr_type(attr)) {
						case ASTDATA_TYPE_STRING:
							strncpy(strbuff, pair->strvalue, sizeof(strbuff) - 1);
							break;

						/* TODO: change output so lvalue comes out as a date */
						case ASTDATA_TYPE_DATE:
						case ASTDATA_TYPE_INTEGER:
							snprintf(strbuff, sizeof(strbuff), "%d", pair->lvalue);
							break;

						case ASTDATA_TYPE_IPADDR:
							ast_inet_ntoa(strbuff, sizeof(strbuff), pair->ipaddr);
							break;
					} /* end switch (type) */

					/* Replace all 'pattern's in 'dest' with 'strbuff' */
					ast_data_strrepl(buff, destsize, dest, pattern, strbuff);
					memcpy(dest, buff, destsize);

					/* look for the next "${" */
					v = strstr(dest, "${");
				} /* end if (valid pair found) */
			} /* end if (valid attribute found) */
		} /* end if (valid ${ } pair found) */
	} /* end while (v) */

	free(buff);
	return(dest);
}
