/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * PostGreSQL Data Resource
 * 
 * Copyright(C) 2004, Rob Gagnon
 * Rob Gagnon <rgagnon@networkip.net>
 *
 * Contributions by:
 *   Steven Sokol <ssokol@sokol-associates.com>  (IAX Friends Code)
 *   Brian Degenhardt <bmd@fourloop.com>  (Memory Leak Fix)
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>
#include <asterisk/config.h>
#include <asterisk/options.h>
#include <asterisk/channel.h>
#include <asterisk/acl.h>
#include <asterisk/pbx.h>
#include <asterisk/data.h>
#include <asterisk/module.h>
#include <asterisk/logger.h>
#include <asterisk/cli.h>
#include "../asterisk.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <libpq-fe.h>

#define TECH_NAME "PostGreSQL"
/* Interface will be registered with this name, and the section
 * in the config file will use this for its parameters
 * ie:  "[pgsql]"
 */
#define CLI_NAME "pgsql"

/* PGSQL stuff */
#define POOL_SIZE_FLOOR				1
#define POOL_SIZE_CEILING			99

#define DEFAULT_INIT_POOL_SIZE	5
#define DEFAULT_MAX_POOL_SIZE		10

/* Default SQL Statements */
#define DEFAULT_SQL_SELECT_EXTENSION \
	"SELECT * FROM extensions " \
	"WHERE context = '${Context}' AND" \
	" (extension = '${Extension}' OR extension LIKE '\\\\_%%') AND" \
	" priority = '${Priority}' " \
	"ORDER BY SUBSTRING(extension FROM 1 FOR 1) ASC," \
	" CHAR_LENGTH(extension) DESC;"

#define DEFAULT_SQL_SELECT_SIP_ACL \
	"SELECT * FROM sipacl WHERE name='${Name}' " \
	"ORDER BY priority;"

#define DEFAULT_SQL_SELECT_SIP_FRIEND_NAME \
	"SELECT * FROM sipfriends WHERE name='${Name}' " \
	"AND (type=${Type} OR type=3);"

#define DEFAULT_SQL_SELECT_SIP_FRIEND_IP \
	"SELECT * FROM sipfriends WHERE ipaddr='${IPAddress}' AND port=${Port} " \
	"AND (type=${Type} OR type=3);"

#define DEFAULT_SQL_UPDATE_SIP_FRIEND \
	"UPDATE sipfriends " \
	"SET ipaddr='${IPAddress}', port=${Port}, regseconds=${RegSeconds}," \
	" username='${Username}', useragent='${UserAgent}'," \
	" fullcontact='${FullContact}' WHERE name='${Name}'"

#define DEFAULT_SQL_SELECT_VM_USER \
	"SELECT * FROM users " \
	"WHERE context='${Context}' AND mailbox='${Mailbox}'"

#define DEFAULT_SQL_SELECT_VM_DIRECTORY_LNAME \
	"SELECT * FROM users " \
	"WHERE context='${Context}' AND fullname SIMILAR TO '%% ${Pattern}%%' " \
	"ORDER BY fullname;"

#define DEFAULT_SQL_SELECT_VM_DIRECTORY_FNAME \
	"SELECT * FROM users " \
	"WHERE context='${Context}' AND fullname SIMILAR TO '${Pattern}%%' " \
	"ORDER BY fullname;"

#define DEFAULT_SQL_UPDATE_VM_PASSWORD_CHANGE \
	"UPDATE users SET password='${NewPassword}' " \
	"WHERE context='${Context}' AND mailbox='${Mailbox}' " \
	"AND password='${Password}'"

#define DEFAULT_SQL_SELECT_IAX_FRIEND_NAME \
	"SELECT * FROM iaxfriends WHERE name='${Name}'" \
	" AND (type=${Type} OR type=3);"

#define DEFAULT_SQL_SELECT_IAX_FRIEND_IP \
	"SELECT * FROM iaxfriends WHERE ipaddr='${IPAddress}' AND port=${Port}" \
	" AND (type=${Type} OR type=3);"

#define DEFAULT_SQL_UPDATE_IAX_FRIEND \
	"UPDATE iaxfriends SET ipaddr='${IPAddress}', port=${Port}, " \
	"regseconds=${RegSeconds} WHERE name='${Name}';"

static char *desc = TECH_NAME " Data Resource";
static int sip_capability = AST_FORMAT_ULAW | AST_FORMAT_ALAW | AST_FORMAT_GSM | AST_FORMAT_H263;
static int iax2_capability = IAX_CAPABILITY_FULLBANDWIDTH;
static int debug_sql = 0;

struct db_pool_entry {
	int							id;
	PGconn						*dbconn;
	ast_mutex_t					lock;
	int							connected;
	time_t						connect_time;
	int							checkedout;
	time_t						checkout_time;
	int							readcount;
	int							writecount;
	int							deletecount;
	struct db_pool_entry		*next;
};

struct db_entry {
	ASTDATA_DBINFO				dbinfo;
	ast_mutex_t					pool_lock;
	struct db_pool_entry		*pool;
	int							initial_dbpool_size;
	int							max_dbpool_size;
	int							pool_size;
	struct db_entry			*next;
};

static struct db_entry *db_list = NULL;

/* CLI stuff */
static char ast_data_status_help[] =
	"Usage: show data "CLI_NAME" status\n"
	"       Shows current connection status for data_"CLI_NAME"\n";

static char ast_data_debug_sql_help[] =
	"Usage: data "CLI_NAME" debug sql\n"
	"       Enables viewing of all SQL statements executing in data_"CLI_NAME"\n";

static char ast_data_no_debug_sql_help[] =
	"Usage: data "CLI_NAME" no debug sql\n"
	"       Disables viewing of all SQL statements executing in data_"CLI_NAME"\n";

/***************************************************************************/
/* Connect to a database.  For other database engines, just change the
 * interior of this function.  Return RESULT_FAILURE if connect fails
 * RESULT_SUCCESS if good
 */
static int
db_connect(ASTDATA_DBINFO *dbinfo, struct db_pool_entry *poolentry) {
	
	if (poolentry == NULL || dbinfo == NULL) {
		return(RESULT_FAILURE);
	}

	if (!dbinfo->options[0]) {
		return(RESULT_FAILURE);
	}
	
	poolentry->dbconn = PQconnectdb(dbinfo->options);

	if (PQstatus(poolentry->dbconn) == CONNECTION_BAD) {
		ast_log(LOG_ERROR,
			"ID%d: Failed to connect to "CLI_NAME
			" database %s@%s. Error from DBEngine: %s\n",
			poolentry->id, PQdb(poolentry->dbconn),
			PQhost(poolentry->dbconn),
			PQerrorMessage(poolentry->dbconn));
		poolentry->connected = 0;
	}
	else {
		ast_log(LOG_DEBUG,"ID%d: Successfully connected to database.\n",
			poolentry->id);
		poolentry->connected = 1;
		poolentry->readcount = 0;
		poolentry->deletecount = 0;
		poolentry->writecount = 0;
		poolentry->connect_time = time(NULL);
	}
	strncpy(dbinfo->hostname, PQhost(poolentry->dbconn),
		sizeof(dbinfo->hostname)-1);
	strncpy(dbinfo->dbname, PQdb(poolentry->dbconn),
		sizeof(dbinfo->dbname)-1);
	dbinfo->port = atoi(PQport(poolentry->dbconn));

	if (poolentry->connected)
		return(RESULT_SUCCESS);
	else
		return(RESULT_FAILURE);
}

/***************************************************************************/
/* Disconnect from the database.  For other databases, just change the
 * interior of this function
 */
static void
db_disconnect(struct db_pool_entry *poolentry) {
	if (poolentry == NULL) {
		return;
	}

	if (poolentry->dbconn) {
		PQfinish(poolentry->dbconn);
		poolentry->dbconn = NULL;
		poolentry->connected = 0;
	}
}

/***************************************************************************/

static void
destroy_dbpools(void) {
	struct db_entry *nextdb = NULL;
	struct db_pool_entry *nextpool = NULL;

	/* walk the db_entries */
	while (db_list) {
		nextdb = db_list->next;
		ast_mutex_lock(&db_list->pool_lock);

		/* walk all the pool entries in this db_entry */
		while (db_list->pool) {
			nextpool = db_list->pool->next;
			db_disconnect(db_list->pool);
			free(db_list->pool);
			db_list->pool = nextpool;
		}
		db_list->pool_size = 0;
		
		/* Free the query info structs for this db_entry */
		ast_data_free_qinfo(&db_list->dbinfo);

		ast_mutex_unlock(&db_list->pool_lock);
		free(db_list);
		db_list = nextdb;
	}

	return;
}

/***************************************************************************/

static struct db_pool_entry *
create_dbpool_entry(struct db_pool_entry **poolhead, int id, int max_pool_size) {
	struct db_pool_entry *newentry = NULL;

	/* make sure we aren't trying to create another entry
	 * outside of the allowed pool size
	 */
	if (id > max_pool_size) {
		/* request to create entry outside of allowed limits */
		ast_log(LOG_WARNING,
			"Creation of DB pool entry outside of MaxPoolSize of %d denied.\n",
			max_pool_size);
	}
	else {
		/* malloc memory for another db pool entry */
		newentry = calloc(1, sizeof(struct db_pool_entry));
		if (newentry == NULL) {
			ast_log(LOG_ERROR, "Out of memory.\n");
		}
		else {
			/* initialize the entry */
			newentry->id = id;
			ast_mutex_init(&newentry->lock);

			/* add entry to the chain at the end */
			if (*poolhead == NULL) {
				*poolhead = newentry;
			}
			else {
				struct db_pool_entry *i = NULL;
				for (i = *poolhead; i->next; i = i->next)
					;
				i->next = newentry;
			}
		}
	} /* end if (dbpool_size check) */

	return(newentry);
}

/***************************************************************************/

static int
initialize_dbpools(void) {
	int i = 0;
	int result = RESULT_SUCCESS;
	struct db_pool_entry *poolentry = NULL;
	struct db_entry *dbentry = NULL;

	/* destroy the existing dbpool(s) */
	/* destroy_dbpools(); */

	dbentry = db_list;
	while (dbentry) {
		/* check that connection pool variables aren't all wacky
	 	* kinda verbose, but clear msgs to the user are important.
	 	*/
		if (dbentry->initial_dbpool_size < POOL_SIZE_FLOOR) {
			ast_log(LOG_WARNING,
				"InitPoolSize less than floor %d in [%s]. Setting to %d.\n",
				POOL_SIZE_FLOOR, CLI_NAME, POOL_SIZE_FLOOR);
			dbentry->initial_dbpool_size = POOL_SIZE_FLOOR;
		}
	
		if (dbentry->initial_dbpool_size > POOL_SIZE_CEILING) {
			ast_log(LOG_WARNING,
				"InitPoolSize greater than ceiling %d in [%s]. Setting to %d.\n",
				POOL_SIZE_CEILING, CLI_NAME, POOL_SIZE_CEILING);
			dbentry->initial_dbpool_size = POOL_SIZE_CEILING;
		}
	
		if (dbentry->max_dbpool_size < dbentry->initial_dbpool_size) {
			ast_log(LOG_WARNING,
				"MaxPoolSize less than InitPoolSize %d in [%s]. Setting to %d.\n",
				dbentry->initial_dbpool_size, CLI_NAME,
				dbentry->initial_dbpool_size);
			dbentry->max_dbpool_size = dbentry->initial_dbpool_size;
		}
	
		if (dbentry->max_dbpool_size > POOL_SIZE_CEILING) {
			ast_log(LOG_WARNING,
				"MaxPoolSize greater than ceiling %d in [%s]. Setting to %d.\n",
				POOL_SIZE_CEILING, CLI_NAME, POOL_SIZE_CEILING);
			dbentry->max_dbpool_size = POOL_SIZE_CEILING;
		}
	
		ast_mutex_lock(&dbentry->pool_lock);

		/* create minimum pool size */
		for (i = 1; i <= dbentry->initial_dbpool_size; i++) {
			poolentry = create_dbpool_entry(&dbentry->pool, i,
								dbentry->max_dbpool_size);

			if (poolentry) {
				dbentry->pool_size++;
				db_connect(&dbentry->dbinfo, poolentry);
			}
		}
		ast_mutex_unlock(&dbentry->pool_lock);

 		dbentry = dbentry->next;
	}

	return(result);
}

/***************************************************************************/

static struct db_pool_entry *
checkout_db_connection(ASTDATA_QTYPE qtype, ASTDATA_QINFO **qinfo) {
	struct db_pool_entry *poolentry = NULL;
	struct db_entry *dbentry = NULL;

	/* find a db pool that can handle this query */
	dbentry = db_list;
	while (dbentry) {
		if ((*qinfo = ast_data_find_qinfo(&dbentry->dbinfo, qtype))) {
			break;
		}
		dbentry = dbentry->next;
	}
	
	if (!dbentry) {
		/* nothing available */
		return(NULL);
	}

	ast_mutex_lock(&dbentry->pool_lock);

	poolentry = dbentry->pool;
	while (poolentry) {
		/* find an entry that is valid, connected, not checked out,
		 */
		if (poolentry->dbconn != NULL &&
			 poolentry->connected == 1 &&
			 poolentry->checkedout == 0) {

			poolentry->checkedout = 1;
			poolentry->checkout_time = time(NULL);
			break;
		}
		poolentry = poolentry->next;
	}

	/* if nothing available in existing db pool,
	 * try creating a new one if not at max allowed.
	 */
	if (poolentry == NULL && (dbentry->pool_size < dbentry->max_dbpool_size)) {
		poolentry = create_dbpool_entry(&dbentry->pool, ++dbentry->pool_size,
							dbentry->max_dbpool_size);
		if (poolentry) {
			if (db_connect(&dbentry->dbinfo, poolentry) == RESULT_SUCCESS) {
				poolentry->checkedout = 1;
				poolentry->checkout_time = time(NULL);
			}
			else {
				poolentry = NULL;
			}
		}
	}

	ast_mutex_unlock(&dbentry->pool_lock);

	return(poolentry);
}

/***************************************************************************/

static void
log_bad_param(ASTDATA_PACKET *packet, ASTDATA_ATTR attribute, char *paramname) {
	char errmsg[80] = "";

	/* log with either the paramname given, or the value of attribute */
	if (paramname) {
		snprintf(errmsg, sizeof(errmsg), "Bad or missing parameter: %s",
			paramname);
	}
	else {
		snprintf(errmsg, sizeof(errmsg), "Missing parameter (%3d)%s",
			attribute, ast_data_attr_name(attribute));
	}

	ast_log(LOG_NOTICE, "%s\n", errmsg);

	ast_data_set_error(packet, ASTDATA_E_BAD_PARAM, errmsg);
}

/***************************************************************************/

static int
show_pool_entry_status(int fd, struct db_pool_entry *poolentry) {
	if (poolentry == NULL) {
		return(RESULT_FAILURE);
	}

	if (poolentry->connected) {
		char in_out[30] = "";
		int ctime = time(NULL) - poolentry->connect_time;
		int days = 0;
		int hours = 0;
		int minutes = 0;
		int seconds = 0;

		days = ctime / 86400;
		ctime -= days*86400;
		hours = ctime / 3600;
		ctime -= hours*3600;
		minutes = ctime / 60;
		ctime -= minutes*60;
		seconds = ctime;

		if (poolentry->checkedout) {
			snprintf(in_out, sizeof(in_out), "Out %d",
				(int)(time(NULL) - poolentry->checkout_time));
		}
		else {
			/* not strncpy cuz its so small */
			strcpy(in_out, "In");
		}

		ast_cli(fd, "%3d\t%s\t%6d\t%6d\t%6d\tConnected %d days, %02d:%02d:%02d\n",
			poolentry->id, in_out,
			poolentry->readcount, poolentry->writecount, poolentry->deletecount,
			days, hours, minutes, seconds);

		return(RESULT_SUCCESS);
	}
	else {
		ast_cli(fd, "%3d\tDisconnected\n", poolentry->id);
		return(RESULT_FAILURE);
	}
}

/***************************************************************************/

static int
show_db_entry_status(int fd, struct db_entry *db) {
	char status[256] = "unknown";
	struct db_pool_entry *poolentry = NULL;
	ASTDATA_QINFO *qi = NULL;
	int result = RESULT_SUCCESS;

	if (!db->dbinfo.options[0]) {
		return(result);
	}

	if (db->dbinfo.port)
		snprintf(status, sizeof(status), "%s@%s:%d",
			db->dbinfo.dbname, db->dbinfo.hostname, db->dbinfo.port);
	else if (db->dbinfo.sock[0])
		snprintf(status, sizeof(status),
			"%s on socket file %s",
			db->dbinfo.dbname, db->dbinfo.sock);
	else
		snprintf(status, sizeof(status), "%s@%s",
			db->dbinfo.dbname, db->dbinfo.hostname);

	ast_cli(fd, "Database: %s\n", status);
	ast_cli(fd, "Queries Handled: \n");
	qi = db->dbinfo.qinfo;
	while (qi) {
		ast_cli(fd, "     %-35s", ast_data_qtype2str(qi->qtype));
		qi = qi->next;
	}
	ast_cli(fd, "\nID\tIn/Out\tReads\tWrites\tDeletes\tStatus\n");
	poolentry = db->pool;
	while (poolentry) {
		result = show_pool_entry_status(fd, poolentry);
		poolentry = poolentry->next;
	}
	ast_cli(fd, "==============================================================================\n");
	return(result);
}

/***************************************************************************/

static int
handle_ast_data_status(int fd, int argc, char *argv[]) {
	struct db_entry *dbentry = NULL;

	ast_cli(fd, "==============================================================================\n");
	dbentry = db_list;
	while (dbentry) {
		show_db_entry_status(fd, dbentry);
		dbentry = dbentry->next;
	}
	return(RESULT_SUCCESS);
}

/***************************************************************************/

static int
handle_ast_data_debug_sql(int fd, int argc, char *argv[]) {
	debug_sql = 1;
	ast_cli(fd, "SQL debug ON for data_"CLI_NAME"\n");
	return(RESULT_SUCCESS);
}

/***************************************************************************/

static int
handle_ast_data_no_debug_sql(int fd, int argc, char *argv[]) {
	debug_sql = 0;
	ast_cli(fd, "SQL debug OFF for data_"CLI_NAME"\n");
	return(RESULT_SUCCESS);
}

/***************************************************************************/

static struct ast_cli_entry ast_data_status_cli =
{ { "show", "data", CLI_NAME, "status", NULL},
		handle_ast_data_status, "Show connection status of data_"CLI_NAME,
		ast_data_status_help, NULL};

static struct ast_cli_entry ast_data_debug_sql_cli =
{ { "data", CLI_NAME, "debug", "sql", NULL},
		handle_ast_data_debug_sql, "Show SQL running in data_"CLI_NAME,
		ast_data_debug_sql_help, NULL};

static struct ast_cli_entry ast_data_no_debug_sql_cli =
{ { "data", CLI_NAME, "no", "debug", "sql", NULL},
		handle_ast_data_no_debug_sql, "Stop showing SQL running in data_"CLI_NAME,
		ast_data_no_debug_sql_help, NULL};

/***************************************************************************/
/* This has been modified by Michael Shuler (mike@bwsys.net) to support
 * pattern matching.
 */
static ASTDATA_PACKET *
pgsql_ext_lookup(ASTDATA_PACKET *request) {
	char *context = NULL;
	char *exten = NULL;
	int priority = 0;
	int action = 0;
	char *esc_context = NULL;
	char *esc_exten = NULL;
	char *dberr = NULL;

	char *dbextension = NULL;
	char *dbapplication = NULL;
	char *dbarguments = NULL;

	ASTDATA_ERROR	errnum = ASTDATA_E_OK;
	ASTDATA_PAIR	*vp = NULL;
	ASTDATA_PACKET	*response = NULL;

	PGconn *dbconn = NULL;
	PGresult *result = NULL;
	char query[512] = "";
	char buff[512] = "";
	char *fname = NULL;
	char *fvalue = NULL;
	int numrows = 0;
	int numfields = 0;
	int rownum = 0;
	int fieldnum = 0;

	struct db_pool_entry *poolentry = NULL;
	ASTDATA_QINFO *qinfo = NULL;

	response = ast_data_create_packet(request->qtype,
		ASTDATA_ATTR_DATASOURCE, CLI_NAME,
		NULL);
	
	if (response == NULL) {
		errno = ENOMEM;
		return(NULL);
	}

	/* If we got this far, the response packet is setup. 
	 * Must return(response) any time past this point, or will
	 * have a memory leak
	 */

	if (request == NULL) {
		/* bad packet */
		log_bad_param(response, ASTDATA_ATTR_INVALID, "invalid request ptr");
		return(response);
	}

	/* get the values from the request packet */
	errnum = ast_data_extract_packet(request,
					ASTDATA_ATTR_CONTEXT, &context,
					ASTDATA_ATTR_EXTENSION, &exten,
					ASTDATA_ATTR_PRIORITY, &priority,
					ASTDATA_ATTR_ACTION, &action,
					NULL);
	
	if (errnum != ASTDATA_E_OK) {
		log_bad_param(response, errnum, NULL);
		return(response);
	}

	if (strlen(exten) >= 128) {
		log_bad_param(response, ASTDATA_ATTR_INVALID, "extension too long");
		return(response);
	}

	poolentry = checkout_db_connection(request->qtype, &qinfo);
	if (poolentry == NULL) {
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database pool available");
		return(response);
	}
	
	dbconn = poolentry->dbconn;
	if (dbconn == NULL) {
		poolentry->checkedout = 0;
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database connection");
		return(response);
	}

	/* parameters are all good at this point */
	esc_context = calloc(1, strlen(context) * 2 + 1);
	esc_exten = calloc(1, strlen(exten) * 2 + 1);
	if (!esc_context || !esc_exten) {
		poolentry->checkedout = 0;
		ast_data_set_error(response, ASTDATA_E_NOMEM, "Out of memory");
		if (esc_context) free(esc_context);
		if (esc_exten) free(esc_exten);
		return(response);
	}

	PQescapeString(esc_context, context, strlen(context));
	PQescapeString(esc_exten, exten, strlen(exten));

	/* Make the query string */
	if (qinfo->query) {
		strncpy(buff, qinfo->query, sizeof(buff) - 1);
	} else {
		strncpy(buff, DEFAULT_SQL_SELECT_EXTENSION, sizeof(buff) - 1);
	}
	ast_data_strrepl(query, sizeof(query), buff, "${Context}", esc_context);
	ast_data_strrepl(buff, sizeof(buff), query, "${Extension}", esc_exten);
	ast_data_var_replace(query, sizeof(query), buff, request);

	free(esc_context);
	free(esc_exten);

	ast_log(LOG_DEBUG, "To match: context=%s, Exten=%s, Priority=%d\n",
		context, exten, priority);

	ast_mutex_lock(&poolentry->lock);

	/* Run the query */
	if (debug_sql) {
		ast_verbose(VERBOSE_PREFIX_2 "data_"CLI_NAME": %s\n", query);
	}
	result = PQexec(dbconn, query);
	if (result == NULL) {
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Query failed: %s\n", dberr);

		/* checkin and unlock pool entry */
		poolentry->checkedout = 0;
		ast_mutex_unlock(&poolentry->lock);
		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}
	poolentry->readcount++;

	/* Get the result set back from the DB */
	if (PQresultStatus(result) == PGRES_BAD_RESPONSE ||
		PQresultStatus(result) == PGRES_NONFATAL_ERROR ||
		PQresultStatus(result) == PGRES_FATAL_ERROR) {
		
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Unable to retrieve query results: %s\n", dberr);

		PQclear(result);
		PQreset(dbconn);
		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);

		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}

	/* See how many rows we got back */
	if ((numrows = PQntuples(result)) == 0) {
		/* Got nothing back, all done. */
		ast_log(LOG_DEBUG, "Found no records\n");

		/* Free the result set memory */
		PQclear(result);

		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);

		ast_data_set_error(response, ASTDATA_E_NOT_FOUND, "No data found");
		return(response);
	}

	ast_log(LOG_DEBUG, "Found %d records\n", numrows);

	/* Compare each row returned from the DB.
	 * NOTE: because we made the ORDER BY ASC anything starting with a '_'
	 * will appear towards the end instead of at the start of the records).
	 * This allows us to apply exact extension matches first and then
	 * pattern matches.
	 */
	numfields = PQnfields(result);
	for (rownum = 0; rownum < numrows; rownum++) {
		for (fieldnum = 0; fieldnum < numfields; fieldnum++) {
			fname = PQfname(result, fieldnum);
			fvalue = PQgetvalue(result, rownum, fieldnum);
			if (fvalue) {
				if (!strcasecmp(fname, "extension")) {
					dbextension = fvalue;
				} else if (!strcasecmp(fname, "application")) {
					dbapplication = fvalue;
				} else if (!strcasecmp(fname, "args")) {
					dbarguments = fvalue;
				}
			} /* end if (fvalue) */
		} /* end for (fieldnum) */

		ast_log(LOG_DEBUG, "Testing for match: %s and %s. Action=%d\n",
			exten, dbextension, action);

		/* Check if we have a match (NOTE: this function also handles
		 * pattern matching)
		 */
		if ((ast_extension_match(dbextension, exten) == 1) ||
			(action == HELPER_CANMATCH)) {

			ast_log(LOG_DEBUG, "Found match: %s and %s\n", exten, dbextension);

			/* Add more data to the response */
			ast_data_add_string(response, ASTDATA_ATTR_EXTENSION, dbextension);
			ast_data_add_string(response, ASTDATA_ATTR_APPLICATION, 
				dbapplication);

			vp = ast_data_add_string(response, ASTDATA_ATTR_ARGUMENTS,
				dbarguments);
			if (vp) {
				/* change commas to pipes in args */
				char *start;
				for (start = vp->strvalue; *start; start++) {
					if (*start == ',')
						*start = '|';
				}
			}

			/* Set the error status to OK on the response packet */
			ast_data_clear_error(response);

			/* All done going through the records, free the memory */
			PQclear(result);

			/* check the db pool entry back in */
			poolentry->checkedout = 0;
			/* Unlock the DB */
			ast_mutex_unlock(&poolentry->lock);

			/* All done, we got our match */
			return(response);
		}

	} /* end for (rownum) */

	/* All done going through the records, free the memory */
	PQclear(result);

	/* Unlock the DB */
	ast_mutex_unlock(&poolentry->lock);
	poolentry->checkedout = 0;

	/* Nothing matched, all done. */
	ast_data_set_error(response, ASTDATA_E_NOT_FOUND, "Not found");
	return(response);
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_sip_update_peer(ASTDATA_PACKET *request) {
	char *peer = NULL;
	char *username = NULL;
	char *useragent = NULL;
	char *fullcontact = NULL;
	char *esc_peer = NULL;
	char *esc_username = NULL;
	char *esc_useragent = NULL;
	char *esc_fullcontact = NULL;
	int regseconds = 0;
	struct in_addr ipaddr = { 0};
	int port = 0;

	char *dberr = NULL;

	ASTDATA_ERROR	errnum = ASTDATA_E_OK;
	ASTDATA_PACKET	*response = NULL;

	PGconn *dbconn = NULL;
	char query[512] = "";
	char buff[512] = "";
	PGresult *result = NULL;

	struct db_pool_entry *poolentry = NULL;
	ASTDATA_QINFO *qinfo = NULL;

	response = ast_data_create_packet(request->qtype,
		ASTDATA_ATTR_DATASOURCE, CLI_NAME,
		NULL);
	
	if (response == NULL) {
		errno = ENOMEM;
		return(NULL);
	}

	/* If we got this far, the response packet is setup.
	 * Must return(response) any time past this point, or will
	 * have a memory leak
	 */

	if (request == NULL) {
		/* bad packet */
		log_bad_param(response, ASTDATA_ATTR_INVALID, "invalid request ptr");
		return(response);
	}

	/* get the values from the request packet */
	errnum = ast_data_extract_packet(request,
					ASTDATA_ATTR_NAME, &peer,
					ASTDATA_ATTR_USERNAME, &username,
					ASTDATA_ATTR_REGSECONDS, &regseconds,
					ASTDATA_ATTR_PORT, &port, 
					ASTDATA_ATTR_IPADDRESS, &ipaddr,
					ASTDATA_ATTR_USERAGENT, &useragent,
					ASTDATA_ATTR_FULLCONTACT, &fullcontact,
					NULL);

	if (errnum != ASTDATA_E_OK) {
		log_bad_param(response, errnum, NULL);
		return(response);
	}

	if (strlen(peer) >= 128) {
		log_bad_param(response, ASTDATA_ATTR_INVALID, "peer too long");
		return(response);
	}

	poolentry = checkout_db_connection(request->qtype, &qinfo);
	if (poolentry == NULL) {
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database pool available");
		return(response);
	}
	
	dbconn = poolentry->dbconn;
	if (dbconn == NULL) {
		poolentry->checkedout = 0;
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database connection");
		return(response);
	}

	/* parameters are all good at this point */
	esc_peer = calloc(1, strlen(peer) * 2 + 1);
	esc_username = calloc(1, strlen(username) * 2 + 1);
	esc_useragent = calloc(1, strlen(useragent) * 2 + 1);
	esc_fullcontact = calloc(1, strlen(fullcontact) * 2 + 1);
	if (!esc_peer || !esc_username || !esc_useragent || !esc_fullcontact) {
		poolentry->checkedout = 0;
		ast_data_set_error(response, ASTDATA_E_NOMEM, "Out of memory");
		if (esc_peer) free(esc_peer);
		if (esc_username) free(esc_username);
		if (esc_useragent) free(esc_useragent);
		if (esc_fullcontact) free(esc_fullcontact);
		return(response);
	}

	PQescapeString(esc_peer, peer, strlen(peer));
	PQescapeString(esc_username, username, strlen(username));
	PQescapeString(esc_useragent, useragent, strlen(useragent));
	PQescapeString(esc_fullcontact, fullcontact, strlen(fullcontact));

	/* Build the query */
	if (qinfo->query) {
		strncpy(buff, qinfo->query, sizeof(buff) - 1);
	} else {
		strncpy(buff, DEFAULT_SQL_UPDATE_SIP_FRIEND, sizeof(buff) - 1);
	}
	ast_data_strrepl(query, sizeof(query), buff, "${Name}", esc_peer);
	ast_data_strrepl(buff, sizeof(buff), query, "${Username}", esc_username);
	ast_data_strrepl(query, sizeof(query), buff, "${UserAgent}", esc_useragent);
	ast_data_strrepl(buff, sizeof(buff), query, "${FullContact}", esc_fullcontact);
	ast_data_var_replace(query, sizeof(query), buff, request);
	
	free(esc_peer);
	free(esc_username);
	free(esc_useragent);
	free(esc_fullcontact);

	ast_mutex_lock(&poolentry->lock);

	/* Run the query */
	if (debug_sql) {
		ast_verbose(VERBOSE_PREFIX_2 "data_"CLI_NAME": %s\n", query);
	}
	result = PQexec(dbconn, query);
	if (result == NULL) {
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_WARNING, "Unable to update database: %s\n", dberr);
		ast_data_set_error(response, ASTDATA_E_UPDATE_FAILED, dberr);
	}
	else {
		PQclear(result);
		poolentry->writecount++;
		ast_data_clear_error(response);
	}

	ast_mutex_unlock(&poolentry->lock);

	poolentry->checkedout = 0;

	return(response);
}

/***************************************************************************/

static int
pgsql_getacl(ASTDATA_PACKET *response, struct db_pool_entry *poolentry) {
	int		sense = 0;
	char		*dberr = NULL;
	char		*username = NULL;
	char		*esc_username = NULL;

	ASTDATA_PAIR	*vp = NULL;

	PGconn	*dbconn = NULL;
	PGresult	*result = NULL;
	char	query[512] = "";
	char	buff[512] = "";
	char	*fname = NULL;
	char	*fvalue = NULL;
	int	numrows = 0;
	int	numfields = 0;
	int	rownum = 0;
	int	fieldnum = 0;

	if (poolentry)	dbconn = poolentry->dbconn;

	vp = ast_data_first_pair(response, ASTDATA_ATTR_NAME);
	if (vp) {
		username = vp->strvalue;
	}

	if (!dbconn || !username || strlen(username) > 128) {
		/* bad params */
		return(0);
	}

	esc_username = calloc(1, strlen(username) * 2 + 1);
	if (!esc_username) {
		ast_data_set_error(response, ASTDATA_E_NOMEM, "Out of memory");
		return(0);
	}

	PQescapeString(esc_username, username, strlen(username));

	/* Make the query */
	strncpy(buff, DEFAULT_SQL_SELECT_SIP_ACL, sizeof(buff) - 1);
	ast_data_strrepl(query, sizeof(query), buff, "${Name}", esc_username);
	
	free(esc_username);

	/* Run the query	*/
	if (debug_sql) {
		ast_verbose(VERBOSE_PREFIX_2 "data_"CLI_NAME": %s\n", query);
	}
	result = PQexec(dbconn, query);
	if (result == NULL) {
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Query failed: %s\n", dberr);

		return(0);
	}
	poolentry->readcount++;

	/* Get the result set back from the DB */
	if (PQresultStatus(result) == PGRES_BAD_RESPONSE ||
		PQresultStatus(result) == PGRES_NONFATAL_ERROR ||
		PQresultStatus(result) == PGRES_FATAL_ERROR) {
		
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Unable to retrieve query results: %s\n", dberr);
		
		PQclear(result);
		PQreset(dbconn);
		
		return(0);
	}

	numrows = PQntuples(result);
	numfields = PQnfields(result);
	for (rownum = 0; rownum < numrows; rownum++) {
		for (fieldnum = 0; fieldnum < numfields; fieldnum++) {
			fname = PQfname(result, fieldnum);
			fvalue = PQgetvalue(result, rownum, fieldnum);
			if (fvalue) {
				if (!strcasecmp(fname, "sense")) {
					if (sscanf(fvalue, "%i", &sense) != 1)
						sense = AST_SENSE_DENY;
					ast_data_add_integer(response, ASTDATA_ATTR_ACLSENSE, sense);
				}
				else if (!strcasecmp(fname, "ipaddr")) {
					ast_data_add_ipaddr(response, ASTDATA_ATTR_ACLIPADDRESS, fvalue);
				}
				else if (!strcasecmp(fname, "netmask")) {
					ast_data_add_ipaddr(response, ASTDATA_ATTR_ACLNETMASK, fvalue);
				}
			}
		} /* end for (fieldnum) */

		/* reset values for next trip around */
		sense = 0;
	} /* end for (rownum) */
	PQclear(result);

	return(numrows);
}

/***************************************************************************/

/*  Used to read a SIP user or peer.  Calling function can just choose not
 *  to use all of the information returned when used for finding a peer.
 */
static ASTDATA_PACKET *
pgsql_sip_find_friend(ASTDATA_PACKET *request, int userpeerflag) {
	int nat = 0;
	int dtmfmode = 0;
	int codecflags = sip_capability;

	char cidname[21] = "";
	char cidnumber[21] = "";
	char cid[48] = "";
	int amaflags = 0;
	int callingpres = 0;
	int canreinvite = 0;

	ASTDATA_ERROR	errnum = ASTDATA_E_OK;
	ASTDATA_PAIR	*vp = NULL;
	ASTDATA_PACKET	*response = NULL;

	char *dberr = NULL;
	char *peer = NULL;
	char *esc_peer = NULL;
	struct in_addr ipaddr = { 0};
	int port = 0;
	time_t regseconds, nowtime;

	PGconn *dbconn = NULL;
	PGresult *result = NULL;
	char query[512] = "";
	char buff[512] = "";
	char numbuff[10] = "";
	char *fname = NULL;
	char *fvalue = NULL;
	int numrows = 0;
	int numfields = 0;
	int fieldnum = 0;

	struct db_pool_entry *poolentry = NULL;
	ASTDATA_QINFO *qinfo = NULL;

	response = ast_data_create_packet(request->qtype,
		ASTDATA_ATTR_DATASOURCE, CLI_NAME,
		NULL);
	
	if (response == NULL) {
		errno = ENOMEM;
		return(NULL);
	}

	/* If we got this far, the response packet is setup.
	 * Must return(response) any time past this point, or will
	 * have a memory leak
	 */

	if (request == NULL) {
		/* not valid pointer coming in */
		log_bad_param(response, ASTDATA_ATTR_INVALID, "invalid request ptr");
		return(response);
	}

	/* see if we got NAME, or (IP and PORT) */
	errnum = ast_data_check_params(request, ASTDATA_ATTR_NAME, NULL);
	if (errnum == ASTDATA_E_OK) {
		/* get the values from the request packet */
		peer = ast_data_first_pair(request, ASTDATA_ATTR_NAME)->strvalue;
	}
	else {
		errnum = ast_data_extract_packet(request,
						ASTDATA_ATTR_IPADDRESS, &ipaddr,
						ASTDATA_ATTR_PORT, &port,
						NULL);
	}

	if (errnum != ASTDATA_E_OK) {
		log_bad_param(response, errnum, NULL);
		return(response);
	}
	
	poolentry = checkout_db_connection(request->qtype, &qinfo);
	if (poolentry == NULL)	{
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database pool available");
		return(response);
	}

	dbconn = poolentry->dbconn;
	if (dbconn == NULL) {
		/* invalid db connection */
		poolentry->checkedout = 0;
		ast_data_set_error(response, ASTDATA_E_NO_DB_CONN, 
			"No database connection");
		return(response);
	}

	snprintf(numbuff, sizeof(numbuff), "%d", userpeerflag);

	if (peer) {
		/* Lookup based on peer name only */
		esc_peer = calloc(1, strlen(peer) * 2 + 1);
		if (!esc_peer) {
			poolentry->checkedout = 0;
			ast_data_set_error(response, ASTDATA_E_NOMEM, "Out of memory");
			return(response);
		}

		PQescapeString(esc_peer, peer, strlen(peer));

		if (qinfo->query) {
			strncpy(query, qinfo->query, sizeof(query) - 1);
		} else {
			strncpy(query, DEFAULT_SQL_SELECT_SIP_FRIEND_NAME, sizeof(query) - 1);
		}
		ast_data_strrepl(buff, sizeof(buff), query, "${Name}", esc_peer);
		ast_data_strrepl(query, sizeof(query), buff, "${Type}", numbuff);

		free(esc_peer);
	}
	else {
		/* Lookup based on IP and PORT combination */
		if (qinfo->altquery) {
			strncpy(query, qinfo->altquery, sizeof(query) - 1);
		} else {
			strncpy(query, DEFAULT_SQL_SELECT_SIP_FRIEND_IP, sizeof(query) - 1);
		}
		ast_data_strrepl(buff, sizeof(buff), query, "${Type}", numbuff);
		ast_data_var_replace(query, sizeof(query), buff, request);
	}

	ast_mutex_lock(&poolentry->lock);

	/* Run the query */
	if (debug_sql) {
		ast_verbose(VERBOSE_PREFIX_2 "data_"CLI_NAME": %s\n", query);
	}
	result = PQexec(dbconn, query);
   if (result == NULL) {
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Query failed: %s\n", dberr);

		/* checkin and unlock pool entry */
		poolentry->checkedout = 0;
		ast_mutex_unlock(&poolentry->lock);
		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}
	poolentry->readcount++;

	/* Get the result set back from the DB */
	if (PQresultStatus(result) == PGRES_BAD_RESPONSE ||
		PQresultStatus(result) == PGRES_NONFATAL_ERROR ||
		PQresultStatus(result) == PGRES_FATAL_ERROR) {

		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Unable to retrieve query results: %s\n", dberr);

		PQclear(result);
		PQreset(dbconn);
		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);

		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}

	/* See how many rows we got back
	 * (In this case, we want only 1)
	 */
	if ((numrows = PQntuples(result)) != 1) {
		/* Got nothing back, all done. */
		ast_log(LOG_DEBUG, "Found %d records\n", numrows);
		
		/* Free the result set memory */
		PQclear(result);
		
		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);
		
		ast_data_set_error(response, ASTDATA_E_NOT_FOUND, "No data found");
		return(response);
	}

	/* Step through the fields */
	numfields = PQnfields(result);
	for (fieldnum = 0; fieldnum < numfields; fieldnum++) {
		fname = PQfname(result, fieldnum);
		fvalue = PQgetvalue(result, 0, fieldnum);
		if (fvalue) {
			if (!strcasecmp(fname, "secret")) {
				ast_data_add_string(response, ASTDATA_ATTR_PASSWORD, fvalue);
			}
			else if (!strcasecmp(fname, "md5secret")) {
				ast_data_add_string(response, ASTDATA_ATTR_MD5PASSWORD, fvalue);
			}
			else if (!strcasecmp(fname, "name")) {
				ast_data_add_string(response, ASTDATA_ATTR_NAME, fvalue);
			}
			else if (!strcasecmp(fname, "context")) {
				ast_data_add_string(response, ASTDATA_ATTR_CONTEXT, fvalue);
			}
			else if (!strcasecmp(fname, "username")) {
				ast_data_add_string(response, ASTDATA_ATTR_USERNAME, fvalue);
			}
			else if (!strcasecmp(fname, "cidname")) {
				strncpy(cidname, fvalue, sizeof(cidname) - 1);
				ast_data_add_string(response, ASTDATA_ATTR_CALLERID_NAME,
					cidname);
			}
			else if (!strcasecmp(fname, "cidnumber")) {
				strncpy(cidnumber, fvalue, sizeof(cidnumber) - 1);
				ast_data_add_string(response, ASTDATA_ATTR_CALLERID_NUM,
					cidnumber);
			}
			else if (!strcasecmp(fname, "accountcode")) {
				ast_data_add_string(response, ASTDATA_ATTR_ACCOUNTCODE, fvalue);
			}
			else if (!strcasecmp(fname, "useragent")) {
				ast_data_add_string(response, ASTDATA_ATTR_USERAGENT, fvalue);
			}
			else if (!strcasecmp(fname, "ipaddr")) {
				ast_data_add_ipaddr(response, ASTDATA_ATTR_IPADDRESS, fvalue);
			}
			else if (!strcasecmp(fname, "mailbox")) {
				ast_data_add_string(response, ASTDATA_ATTR_MAILBOX, fvalue);
			}
			else if (!strcasecmp(fname, "fullcontact")) {
				ast_data_add_string(response, ASTDATA_ATTR_FULLCONTACT, fvalue);
			}
			else if (!strcasecmp(fname, "language")) {
				ast_data_add_string(response, ASTDATA_ATTR_LANGUAGE, fvalue);
			}
			else if (!strcasecmp(fname, "port")) {
				if (sscanf(fvalue, "%i", &port) != 1)
					port = 0;
				ast_data_add_integer(response, ASTDATA_ATTR_PORT, port);
			}
			else if (!strcasecmp(fname, "nat")) {
				sscanf(fvalue, "%i", &nat);
				ast_data_add_integer(response, ASTDATA_ATTR_NATFLAG, nat);
			}
			else if (!strcasecmp(fname, "dtmfmode")) {
				sscanf(fvalue, "%i", &dtmfmode);
				ast_data_add_integer(response, ASTDATA_ATTR_DTMFMODE,
					dtmfmode);
			}
			else if (!strcasecmp(fname, "regseconds")) {
				if (sscanf(fvalue, "%li", &regseconds) != 1)
					regseconds = 0;
				ast_data_add_integer(response, ASTDATA_ATTR_REGSECONDS,
					regseconds);
			}
			else if (!strcasecmp(fname, "codecflags")) {
				if (sscanf(fvalue, "%i", &codecflags) != 1)
					codecflags = sip_capability;
				ast_data_add_integer(response, ASTDATA_ATTR_CODECFLAGS,
					codecflags);
				if (codecflags == 0) {
					ast_log(LOG_WARNING,
						"codecflags for peer: %s set to zero (none)\n", peer);
				}
			}
			else if (!strcasecmp(fname, "canreinvite")) {
				if (sscanf(fvalue, "%i", &canreinvite) != 1)
					canreinvite = 0;
				ast_data_add_integer(response, ASTDATA_ATTR_CANREINVITE,
					canreinvite);
			}
			else if (!strcasecmp(fname, "callingpres")) {
				if (sscanf(fvalue, "%i", &callingpres) != 1)
					callingpres = 0;
				ast_data_add_integer(response, ASTDATA_ATTR_CALLINGPRES,
					callingpres);
			}
			else if (!strcasecmp(fname, "amaflags")) {
				if (sscanf(fvalue, "%i", &amaflags) != 1)
					amaflags = 0;
				ast_data_add_integer(response, ASTDATA_ATTR_AMAFLAGS,
					amaflags);
			}
		} /* end if (fvalue) */
	} /* end for(fieldnum) */

	/* check for expired registration */
	time(&nowtime);
	if (nowtime > regseconds) {
		vp = ast_data_first_pair(response, ASTDATA_ATTR_IPADDRESS);
		if (!vp) {
			/* This should not happen, but just in case.
			 * If you get this, you need ipaddr VARCHAR(20) in your
			 * sipfriends table
			 */
			ast_log(LOG_WARNING,
				"'ipaddr' field missing in sipfriends table.\n");
		}
		else {
			inet_aton("0.0.0.0", &vp->ipaddr);
		}
	}
	PQclear(result);

	/* format caller id string */
	if (cidname[0]) {
		snprintf(cid, sizeof(cid), "\"%s\" <%s>", cidname, cidnumber);
		ast_data_add_string(response, ASTDATA_ATTR_CALLERID, cid);
	}
	else {
		if (cidnumber[0]) {
			ast_data_add_string(response, ASTDATA_ATTR_CALLERID, cidnumber);
		}
	}

	/* go get the acl if there is one */
	pgsql_getacl(response, poolentry);

	ast_data_clear_error(response);

	poolentry->checkedout = 0;
	ast_mutex_unlock(&poolentry->lock);

	return(response);
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_sip_find_user(ASTDATA_PACKET *request) {
	return(pgsql_sip_find_friend(request, SIP_TYPE_USER));
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_sip_find_peer(ASTDATA_PACKET *request) {
	return(pgsql_sip_find_friend(request, SIP_TYPE_PEER));
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_vm_find_user(ASTDATA_PACKET *request) {
	char *context = NULL;
	char *mailbox = NULL;
	char *esc_context = NULL;
	char *esc_mailbox = NULL;

	ASTDATA_ERROR	errnum = ASTDATA_E_OK;
	ASTDATA_PACKET *response = NULL;

	PGconn *dbconn = NULL;
	PGresult	*result = NULL;
	char *dberr = NULL;
	char query[512] = "";
	char buff[512] = "";
	char *fname = NULL;
	char *fvalue = NULL;
	int numrows = 0;
	int numfields = 0;
	int fieldnum = 0;

	struct db_pool_entry *poolentry = NULL;
	ASTDATA_QINFO *qinfo = NULL;

	response = ast_data_create_packet(request->qtype,
		ASTDATA_ATTR_DATASOURCE, CLI_NAME,
		NULL);
		
	if (response == NULL) {
		errno = ENOMEM;
		return(NULL);
	}

	if (request == NULL) {
		log_bad_param(response, ASTDATA_ATTR_INVALID, "invalid request ptr");
		return(response);
	}

	/* get the values from the request packet */
	errnum = ast_data_extract_packet(request,
					ASTDATA_ATTR_CONTEXT, &context,
					ASTDATA_ATTR_MAILBOX, &mailbox,
					NULL);
					
	if (errnum == ASTDATA_ATTR_CONTEXT) { 
		context = "default"; 
	} 
	else if (errnum != ASTDATA_E_OK) { 
		log_bad_param(response, errnum, NULL); 
		return(response); 
	} 	
	if (context == NULL || context[0] == '\0') { 
		context = "default"; 
	}
	
	poolentry = checkout_db_connection(request->qtype, &qinfo);
	if (poolentry == NULL) {
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database pool available");
		return(response);
	}
	
	dbconn = poolentry->dbconn;
	if (dbconn == NULL) {
		poolentry->checkedout = 0;
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database connection");
		return(response);
	}

	ast_data_add_string(response, ASTDATA_ATTR_CONTEXT, context);
	ast_data_add_string(response, ASTDATA_ATTR_MAILBOX, mailbox);

	esc_context = calloc(1, strlen(context) * 2 + 1);
	esc_mailbox = calloc(1, strlen(mailbox) * 2 + 1);
	if (!esc_context || !esc_mailbox) {
		poolentry->checkedout = 0;
		ast_data_set_error(response, ASTDATA_E_NOMEM, "Out of memory");
		if (esc_context) free(esc_context);
		if (esc_mailbox) free(esc_mailbox);
		return(response);
	}

	PQescapeString(esc_context, context, strlen(context));
	PQescapeString(esc_mailbox, mailbox, strlen(mailbox));

	/* Build the Query */
	if (qinfo->query) {
		strncpy(query, qinfo->query, sizeof(query) - 1);
	} else {
		strncpy(query, DEFAULT_SQL_SELECT_VM_USER, sizeof(query) - 1);
	}
	ast_data_strrepl(buff, sizeof(buff), query, "${Context}", esc_context);
	ast_data_strrepl(query, sizeof(query), buff, "${Mailbox}", esc_mailbox);

	free(esc_context);
	free(esc_mailbox);
	
	ast_mutex_lock(&poolentry->lock);

	/* Run the query */
	if (debug_sql) {
		ast_verbose(VERBOSE_PREFIX_2 "data_"CLI_NAME": %s\n", query);
	}
	result = PQexec(dbconn, query);
	if (result == NULL) {
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Query failed: %s\n", dberr);
		
		/* checkin and unlock pool entry */
		poolentry->checkedout = 0;
		ast_mutex_unlock(&poolentry->lock);
		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}
	poolentry->readcount++;
	
	/* Get the result set back from the DB */
	if (PQresultStatus(result) == PGRES_BAD_RESPONSE ||
		PQresultStatus(result) == PGRES_NONFATAL_ERROR ||
		PQresultStatus(result) == PGRES_FATAL_ERROR) {
		
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Unable to retrieve query results: %s\n", dberr);
		
		PQclear(result);
		PQreset(dbconn);
		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);
		
		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}

	/* See how many rows we got back
	 * (In this case, we want only 1)
	 */
	if ((numrows = PQntuples(result)) != 1) {
		/* Got nothing back, all done. */
		ast_log(LOG_DEBUG, "Found %d records\n", numrows);
		
		/* Free the result set memory */
		PQclear(result);
		
		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);
		
		ast_data_set_error(response, ASTDATA_E_NOT_FOUND, "No data found");
		return(response);
	}

	/* Step through the fields and fields */
	numfields = PQnfields(result);
	for (fieldnum = 0; fieldnum < numfields; fieldnum++) {
		fname = PQfname(result, fieldnum);
		fvalue = PQgetvalue(result, 0, fieldnum);
		if (fvalue) {
			if (!strcmp(fname, "password")) {
				ast_data_add_string(response, ASTDATA_ATTR_PASSWORD, fvalue);
			}
			else if (!strcmp(fname, "fullname")) {
				ast_data_add_string(response, ASTDATA_ATTR_FULLNAME, fvalue);
			}
			else if (!strcmp(fname, "email")) {
				ast_data_add_string(response, ASTDATA_ATTR_EMAIL, fvalue);
			}
			else if (!strcmp(fname, "pager")) {
				ast_data_add_string(response, ASTDATA_ATTR_PAGER, fvalue);
			}
			else if (!strcmp(fname, "options")) {
				ast_data_add_string(response, ASTDATA_ATTR_OPTIONS, fvalue);
			}
		} /* end if (fvalue) */
	} /* end for (fielnum) */

	PQclear(result);

	poolentry->checkedout = 0;
	ast_mutex_unlock(&poolentry->lock);
	ast_data_clear_error(response);

	return(response);
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_vm_directory(ASTDATA_PACKET *request) {
	char *context = NULL;
	char *userinput = NULL;
	char *esc_context = NULL;
	int lastnameflag = 1;
	
	/* for 3 digit userinput, 52 bytes should be the minimum */
	char regexp[80] = "";
	char *esc_regexp = NULL;

	ASTDATA_ERROR	errnum = ASTDATA_E_OK;
	ASTDATA_PACKET *response = NULL;

	PGconn *dbconn = NULL;
	PGresult	*result = NULL;
	char *dberr = NULL;
	char query[512] = "";
	char buff[512] = "";
	char *fname = NULL;
	char *fvalue = NULL;
	int numrows = 0;
	int rownum = 0;
	int numfields = 0;
	int fieldnum = 0;

	struct db_pool_entry *poolentry = NULL;
	ASTDATA_QINFO *qinfo = NULL;

	response = ast_data_create_packet(request->qtype,
		ASTDATA_ATTR_DATASOURCE, CLI_NAME,
		NULL);
		
	if (response == NULL) {
		errno = ENOMEM;
		return(NULL);
	}

	if (request == NULL) {
		log_bad_param(response, ASTDATA_ATTR_INVALID, "invalid request ptr");
		return(response);
	}

	/* get the values from the request packet */
	errnum = ast_data_extract_packet(request,
					ASTDATA_ATTR_CONTEXT, &context,
					ASTDATA_ATTR_USERINPUT, &userinput,
					ASTDATA_ATTR_LASTNAMEFLAG, &lastnameflag,
					NULL);
					
	if (errnum != ASTDATA_E_OK) { 
		log_bad_param(response, errnum, NULL); 
		return(response); 
	} 	
	
	poolentry = checkout_db_connection(request->qtype, &qinfo);
	if (poolentry == NULL) {
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database pool available");
		return(response);
	}
	
	dbconn = poolentry->dbconn;
	if (dbconn == NULL) {
		poolentry->checkedout = 0;
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database connection");
		return(response);
	}

	ast_data_add_string(response, ASTDATA_ATTR_CONTEXT, context);

	/* convert the 3-digit userinput to a regular expression of letters
	 * to match to the database
	 */
	ast_data_dtmf2regexp(regexp, sizeof(regexp), userinput);

	esc_context = calloc(1, strlen(context) * 2 + 1);
	esc_regexp = calloc(1, strlen(regexp) * 2 + 1);
	if (!esc_context || !esc_regexp) {
		poolentry->checkedout = 0;
		ast_data_set_error(response, ASTDATA_E_NOMEM, "Out of memory");
		if (esc_context) free(esc_context);
		if (esc_regexp) free(esc_regexp);
		return(response);
	}

	PQescapeString(esc_context, context, strlen(context));
	PQescapeString(esc_regexp, regexp, strlen(regexp));

	/* Build the Query */
	if (lastnameflag) {
		if (qinfo->query) {
			strncpy(query, qinfo->query, sizeof(query) - 1);
		} else {
			strncpy(query, DEFAULT_SQL_SELECT_VM_DIRECTORY_LNAME, sizeof(query) - 1);
		}
	}
	else {
		/* by first name */
		if (qinfo->altquery) {
			strncpy(query, qinfo->altquery, sizeof(query) - 1);
		} else {
			strncpy(query, DEFAULT_SQL_SELECT_VM_DIRECTORY_FNAME, sizeof(query) - 1);
		}
	}
	ast_data_strrepl(buff, sizeof(buff), query, "${Context}", esc_context);
	ast_data_strrepl(query, sizeof(query), buff, "${Pattern}", esc_regexp);

	free(esc_context);
	free(esc_regexp);
	
	ast_mutex_lock(&poolentry->lock);

	/* Run the query */
	if (debug_sql) {
		ast_verbose(VERBOSE_PREFIX_2 "data_"CLI_NAME": %s\n", query);
	}
	result = PQexec(dbconn, query);
	if (result == NULL) {
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Query failed: %s\n", dberr);
		
		/* checkin and unlock pool entry */
		poolentry->checkedout = 0;
		ast_mutex_unlock(&poolentry->lock);
		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}
	poolentry->readcount++;
	
	/* Get the result set back from the DB */
	if (PQresultStatus(result) == PGRES_BAD_RESPONSE ||
		PQresultStatus(result) == PGRES_NONFATAL_ERROR ||
		PQresultStatus(result) == PGRES_FATAL_ERROR) {
		
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Unable to retrieve query results: %s\n", dberr);
		
		PQclear(result);
		PQreset(dbconn);
		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);
		
		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}

	/* See how many rows we got back
	 * (In this case, we want at least 1)
	 */
	if ((numrows = PQntuples(result)) <= 0) {
		/* Got nothing back, all done. */
		ast_log(LOG_DEBUG, "Found %d records\n", numrows);
		
		/* Free the result set memory */
		PQclear(result);
		
		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);
		
		ast_data_set_error(response, ASTDATA_E_NOT_FOUND, "No data found");
		return(response);
	}

	/* Step through the rosws and fields */
	numfields = PQnfields(result);
	for (rownum = 0; rownum < numrows; rownum++) {
		for (fieldnum = 0; fieldnum < numfields; fieldnum++) {
			fname = PQfname(result, fieldnum);
			fvalue = PQgetvalue(result, rownum, fieldnum);
			if (fvalue) {
				if (!strcmp(fname, "fullname")) {
					ast_data_add_string(response, ASTDATA_ATTR_FULLNAME, fvalue);
				}
				else if (!strcmp(fname, "mailbox")) {
					ast_data_add_string(response, ASTDATA_ATTR_MAILBOX, fvalue);
				}
			} /* end if (fvalue) */
		} /* end for (fielnum) */
	} /* end for (rownum) */

	PQclear(result);

	poolentry->checkedout = 0;
	ast_mutex_unlock(&poolentry->lock);
	ast_data_clear_error(response);

	return(response);
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_vm_change_password(ASTDATA_PACKET *request) {
	char *context = NULL;
	char *mailbox = NULL;
	char *oldpassword = NULL;
	char *newpassword = NULL;
	char *esc_context = NULL;
	char *esc_mailbox = NULL;
	char *esc_oldpassword = NULL;
	char *esc_newpassword = NULL;

	ASTDATA_ERROR	errnum = ASTDATA_E_OK;
	ASTDATA_PACKET	*response = NULL;

	PGconn *dbconn = NULL;
	PGresult *result = NULL;
	char query[512] = "";
	char buff[512] = "";

	struct db_pool_entry *poolentry = NULL;
	ASTDATA_QINFO *qinfo = NULL;

	response = ast_data_create_packet(request->qtype,
		ASTDATA_ATTR_DATASOURCE, CLI_NAME,
		NULL);
	
	if (response == NULL) {
		errno = ENOMEM;
		return(NULL);
	}

	/* If we got this far, the response packet is setup.
	 * Must return(response) any time past this point, or will
	 * have a memory leak
	 */

	if (request == NULL) {
		ast_log(LOG_ERROR, "invalid request pointer.\n");
		return(response);
	}

	/* get the values from the request packet */
	errnum = ast_data_extract_packet(request,
					ASTDATA_ATTR_CONTEXT, &context,
					ASTDATA_ATTR_MAILBOX, &mailbox,
					ASTDATA_ATTR_PASSWORD, &oldpassword,
					ASTDATA_ATTR_NEWPASSWORD, &newpassword,
					NULL);
	
	if (errnum == ASTDATA_ATTR_CONTEXT) {
		context = "default";
	}
	else if (errnum != ASTDATA_E_OK) {
		log_bad_param(response, errnum, NULL);
		return(response);
	}

	if (context == NULL || context[0] == '\0') {
		context = "default";
	}

	poolentry = checkout_db_connection(request->qtype, &qinfo);
	if (poolentry == NULL) {
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database pool available");
		return(response);
	}
	
	dbconn = poolentry->dbconn;
	if (dbconn == NULL) {
		poolentry->checkedout = 0;
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database connection");
		return(response);
	}

	esc_newpassword = calloc(1, strlen(newpassword) * 2 + 1);
	esc_context = calloc(1, strlen(context) * 2 + 1);
	esc_mailbox = calloc(1, strlen(mailbox) * 2 + 1);
	esc_oldpassword = calloc(1, strlen(oldpassword) * 2 + 1);
	if (!esc_newpassword || !esc_context || !esc_mailbox || !esc_oldpassword) {
		poolentry->checkedout = 0;
		ast_data_set_error(response, ASTDATA_E_NOMEM, "Out of memory");
		if (esc_newpassword) free(esc_newpassword);
		if (esc_context) free(esc_context);
		if (esc_mailbox) free(esc_mailbox);
		if (esc_oldpassword) free(esc_oldpassword);
		return(response);
	}

	PQescapeString(esc_newpassword, newpassword, strlen(newpassword));
	PQescapeString(esc_context, context, strlen(context));
	PQescapeString(esc_mailbox, mailbox, strlen(mailbox));
	PQescapeString(esc_oldpassword, oldpassword, strlen(oldpassword));

	/* Build the Query */
	if (qinfo->query) {
		strncpy(query, qinfo->query, sizeof(query));
	} else {
		strncpy(query, DEFAULT_SQL_UPDATE_VM_PASSWORD_CHANGE, sizeof(query));
	}
	ast_data_strrepl(buff, sizeof(buff), query,
		"${NewPassword}", esc_newpassword);
	ast_data_strrepl(query, sizeof(query), buff, "${Context}", esc_context);
	ast_data_strrepl(buff, sizeof(buff), query, "${Mailbox}", esc_mailbox);
	ast_data_strrepl(query, sizeof(query ), buff,
		"${Password}", esc_oldpassword);

	free(esc_context);
	free(esc_mailbox);
	free(esc_oldpassword);
	free(esc_newpassword);

	ast_mutex_lock(&poolentry->lock);

	/* Run the query	*/
	if (debug_sql) {
		ast_verbose(VERBOSE_PREFIX_2 "data_"CLI_NAME": %s\n", query);
	}
	result = PQexec(dbconn, query);
	if (result == NULL) {
		ast_log(LOG_WARNING, "Unable to update database\n");
		ast_data_set_error(response, ASTDATA_E_UPDATE_FAILED,
			"Unable to change password");
	}
	else {
		PQclear(result);
		poolentry->writecount++;
		ast_data_clear_error(response);
	}

	poolentry->checkedout = 0;
	ast_mutex_unlock(&poolentry->lock);

	return(response);
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_iax_update_peer(ASTDATA_PACKET *request) {
	char *peer = NULL;
	char *esc_peer = NULL;
	int regseconds = 0;
	struct in_addr ipaddr = { 0};
	int port = 0;
	time_t nowtime;

	char *dberr = NULL;

	ASTDATA_ERROR	errnum = ASTDATA_E_OK;
	ASTDATA_PACKET	*response = NULL;

	PGconn *dbconn = NULL;
	PGresult *result = NULL;
	char query[512] = "";
	char buff[512] = "";

	struct db_pool_entry *poolentry = NULL;
	ASTDATA_QINFO *qinfo = NULL;

	response = ast_data_create_packet(request->qtype,
		ASTDATA_ATTR_DATASOURCE, CLI_NAME,
		NULL);
	
	if (response == NULL) {
		errno = ENOMEM;
		return(NULL);
	}

	/* If we got this far, the response packet is setup.
	 * Must return(response) any time past this point, or will
	 * have a memory leak
	 */

	if (request == NULL) {
		/* bad packet */
		log_bad_param(response, ASTDATA_ATTR_INVALID, "invalid request ptr");
		return(response);
	}

	/* get the values from the request packet */
	errnum = ast_data_extract_packet(request,
					ASTDATA_ATTR_NAME, &peer,
					ASTDATA_ATTR_IPADDRESS, &ipaddr,
					ASTDATA_ATTR_PORT, &port, 
					ASTDATA_ATTR_REGSECONDS, &regseconds,
					NULL);

	if (errnum != ASTDATA_E_OK) {
		log_bad_param(response, errnum, NULL);
		return(response);
	}

	if (strlen(peer) >= 128) {
		log_bad_param(response, ASTDATA_ATTR_INVALID, "peer too long");
		return(response);
	}

	poolentry = checkout_db_connection(request->qtype, &qinfo);
	if (poolentry == NULL) {
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database pool available");
		return(response);
	}
	
	dbconn = poolentry->dbconn;
	if (dbconn == NULL) {
		poolentry->checkedout = 0;
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database connection");
		return(response);
	}

	/* parameters are all good at this point */
	time(&nowtime);

	esc_peer = calloc(1, strlen(peer) * 2 + 1);
	if (!esc_peer) {
		poolentry->checkedout = 0;
		ast_data_set_error(response, ASTDATA_E_NOMEM, "Out of memory");
		return(response);
	}

	PQescapeString(esc_peer, peer, strlen(peer));

	/* Build the Query */
	if (qinfo->query) {
		strncpy(query, qinfo->query, sizeof(query) - 1);
	} else {
		strncpy(query, DEFAULT_SQL_UPDATE_IAX_FRIEND, sizeof(query) - 1);
	}
	ast_data_strrepl(buff, sizeof(buff), query, "${Name}", esc_peer);
	ast_data_var_replace(query, sizeof(query), buff, request);

	free(esc_peer);

	ast_mutex_lock(&poolentry->lock);

	/* Run the query	*/
	if (debug_sql) {
		ast_verbose(VERBOSE_PREFIX_2 "data_"CLI_NAME": %s\n", query);
	}
	result = PQexec(dbconn, query);
	if (result == NULL) {
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_WARNING, "Unable to update database: %s\n", dberr);
		ast_data_set_error(response, ASTDATA_E_UPDATE_FAILED, dberr);
	}
	else {
		PQclear(result);
		poolentry->writecount++;
		ast_data_clear_error(response);
	}

	ast_mutex_unlock(&poolentry->lock);

	poolentry->checkedout = 0;

	return(response);
}

/***************************************************************************/

/*  Used to read a IAX user or peer.  Calling function can just choose not
 *  to use all of the information returned when used for finding a peer.
 */
static ASTDATA_PACKET *
pgsql_iax_find_friend(ASTDATA_PACKET *request, int userpeerflag) {
	
	int codecflags = iax2_capability;
	char cidname[21] = "";
	char cidnumber[21] = "";
	char cid[48] = "";
	int amaflags = 0;
	int notransfer = 0;

	ASTDATA_ERROR	errnum = ASTDATA_E_OK;
	ASTDATA_PACKET	*response = NULL;

	char *dberr = NULL;
	char *peer = NULL;
	char *esc_peer = NULL;
	struct in_addr ipaddr = { 0};
	int port = 0;
	time_t regseconds;

	PGconn *dbconn = NULL;
	PGresult *result = NULL;
	char query[512] = "";
	char buff[512] = "";
	char numbuff[10] = "";
	char *fname = NULL;
	char *fvalue = NULL;
	int numrows = 0;
	int numfields = 0;
	int fieldnum = 0;

	struct db_pool_entry *poolentry = NULL;
	ASTDATA_QINFO *qinfo = NULL;

	response = ast_data_create_packet(request->qtype,
		ASTDATA_ATTR_DATASOURCE, CLI_NAME,
		NULL);
	
	if (response == NULL) {
		errno = ENOMEM;
		return(NULL);
	}

	/* If we got this far, the response packet is setup.
	 * Must return(response) any time past this point, or will
	 * have a memory leak
	 */

	if (request == NULL) {
		/* not valid pointer coming in */
		log_bad_param(response, ASTDATA_ATTR_INVALID, "invalid request ptr");
		return(response);
	}

	/* see if we got NAME, or (IP and PORT) */
	errnum = ast_data_check_params(request, ASTDATA_ATTR_NAME, NULL);
	if (errnum == ASTDATA_E_OK) {
		/* get the values from the request packet */
		peer = ast_data_first_pair(request, ASTDATA_ATTR_NAME)->strvalue;
	}
	else {
		errnum = ast_data_extract_packet(request,
						ASTDATA_ATTR_IPADDRESS, &ipaddr,
						ASTDATA_ATTR_PORT, &port,
						NULL);
	}

	if (errnum != ASTDATA_E_OK) {
		log_bad_param(response, errnum, NULL);
		return(response);
	}
	
	poolentry = checkout_db_connection(request->qtype, &qinfo);
	if (poolentry == NULL)	{
		ast_data_set_error(response,
			ASTDATA_E_NO_DB_CONN, "No database pool available");
		return(response);
	}

	dbconn = poolentry->dbconn;
	if (dbconn == NULL) {
		/* invalid db connection */
		poolentry->checkedout = 0;
		ast_data_set_error(response, ASTDATA_E_NO_DB_CONN, 
			"No database connection");
		return(response);
	}

	snprintf(numbuff, sizeof(numbuff), "%d", userpeerflag);

	if (peer) {
		/* Lookup based on peer name only */
		esc_peer = calloc(1, strlen(peer) * 2 + 1);
		if (!esc_peer) {
			poolentry->checkedout = 0;
			ast_data_set_error(response, ASTDATA_E_NOMEM, "Out of memory");
			return(response);
		}

		PQescapeString(esc_peer, peer, strlen(peer));

		if (qinfo->query) {
			strncpy(query, qinfo->query, sizeof(query) - 1);
		} else {
			strncpy(query, DEFAULT_SQL_SELECT_IAX_FRIEND_NAME, sizeof(query) - 1);
		}
		ast_data_strrepl(buff, sizeof(buff), query, "${Type}", numbuff);
		ast_data_strrepl(query, sizeof(query), buff, "${Name}", esc_peer);

		free(esc_peer);
	}
	else {
		/* Lookup based on IP and PORT combination */
		if (qinfo->altquery) {
			strncpy(query, qinfo->altquery, sizeof(query) - 1);
		} else {
			strncpy(query, DEFAULT_SQL_SELECT_IAX_FRIEND_IP, sizeof(query) - 1);
		}
		ast_data_strrepl(buff, sizeof(buff), query, "${Type}", numbuff);
		ast_data_var_replace(query, sizeof(query), buff, request);
	}

	ast_mutex_lock(&poolentry->lock);

	/* Run the query */
	if (debug_sql) {
		ast_verbose(VERBOSE_PREFIX_2 "data_"CLI_NAME": %s\n", query);
	}
	result = PQexec(dbconn, query);
   if (result == NULL) {
		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Query failed: %s\n", dberr);

		/* checkin and unlock pool entry */
		poolentry->checkedout = 0;
		ast_mutex_unlock(&poolentry->lock);
		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}
	poolentry->readcount++;

	/* Get the result set back from the DB */
	if (PQresultStatus(result) == PGRES_BAD_RESPONSE ||
		PQresultStatus(result) == PGRES_NONFATAL_ERROR ||
		PQresultStatus(result) == PGRES_FATAL_ERROR) {

		dberr = (char *)PQerrorMessage(dbconn);
		ast_log(LOG_ERROR, "Unable to retrieve query results: %s\n", dberr);

		PQclear(result);
		PQreset(dbconn);
		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);

		ast_data_set_error(response, ASTDATA_E_FIND_FAILED, dberr);
		return(response);
	}

	/* See how many rows we got back
	 * (In this case, we want only 1)
	 */
	if ((numrows = PQntuples(result)) != 1) {
		/* Got nothing back, all done. */
		ast_log(LOG_DEBUG, "Found %d records\n", numrows);
		
		/* Free the result set memory */
		PQclear(result);
		
		poolentry->checkedout = 0;
		/* Unlock the DB */
		ast_mutex_unlock(&poolentry->lock);
		
		ast_data_set_error(response, ASTDATA_E_NOT_FOUND, "No data found");
		return(response);
	}

	/* Step through the fields */
	numfields = PQnfields(result);
	for (fieldnum = 0; fieldnum < numfields; fieldnum++) {
		fname = PQfname(result, fieldnum);
		fvalue = PQgetvalue(result, 0, fieldnum);
		if (fvalue) {
			if (!strcasecmp(fname, "secret")) {
				ast_data_add_string(response, ASTDATA_ATTR_PASSWORD, fvalue);
			}
			else if (!strcasecmp(fname, "name")) {
				ast_data_add_string(response, ASTDATA_ATTR_NAME, fvalue);
			}
			else if (!strcasecmp(fname, "context")) {
				ast_data_add_string(response, ASTDATA_ATTR_CONTEXT, fvalue);
			}
			else if (!strcasecmp(fname, "cidname")) {
				strncpy(cidname, fvalue, sizeof(cidname) - 1);
				ast_data_add_string(response, ASTDATA_ATTR_CALLERID_NAME,
					cidname);
			}
			else if (!strcasecmp(fname, "cidnumber")) {
				strncpy(cidnumber, fvalue, sizeof(cidnumber) - 1);
				ast_data_add_string(response, ASTDATA_ATTR_CALLERID_NUM,
					cidnumber);
			}
			else if (!strcasecmp(fname, "accountcode")) {
				ast_data_add_string(response, ASTDATA_ATTR_ACCOUNTCODE, fvalue);
			}
			else if (!strcasecmp(fname, "ipaddr")) {
				ast_data_add_ipaddr(response, ASTDATA_ATTR_IPADDRESS, fvalue);
			}
			else if (!strcasecmp(fname, "mailbox")) {
				ast_data_add_string(response, ASTDATA_ATTR_MAILBOX, fvalue);
			}
			else if (!strcasecmp(fname, "port")) {
				if (sscanf(fvalue, "%i", &port) != 1)
					port = 0;
				ast_data_add_integer(response, ASTDATA_ATTR_PORT,
					port);
			}
			else if (!strcasecmp(fname, "regseconds")) {
				if (sscanf(fvalue, "%li", &regseconds) != 1)
					regseconds = 0;
				ast_data_add_integer(response, ASTDATA_ATTR_REGSECONDS,
					regseconds);
			}
			else if (!strcasecmp(fname, "codecflags")) {
				if (sscanf(fvalue, "%i", &codecflags) != 1)
					codecflags = sip_capability;
				ast_data_add_integer(response, ASTDATA_ATTR_CODECFLAGS,
					codecflags);
				if (codecflags == 0) {
					ast_log(LOG_WARNING,
						"codecflags for peer: %s set to zero (none)\n", peer);
				}
			}
			else if (!strcasecmp(fname, "inkeys")) {
				ast_data_add_string(response, ASTDATA_ATTR_PUBLICKEY, fvalue);
			}
			else if (!strcasecmp(fname, "outkey")) {
				ast_data_add_string(response, ASTDATA_ATTR_PRIVATEKEY, fvalue);
			}
			else if (!strcasecmp(fname, "notransfer")) {
				if (sscanf(fvalue, "%i", &notransfer) != 1)
					notransfer = 0;
				ast_data_add_integer(response, ASTDATA_ATTR_NOTRANSFERFLAG,
					notransfer);
			}
			else if (!strcasecmp(fname, "amaflags")) {
				if (sscanf(fvalue, "%i", &amaflags) != 1)
					amaflags = 0;
				ast_data_add_integer(response, ASTDATA_ATTR_AMAFLAGS,
					amaflags);
			}
		} /* end if (fvalue) */
	} /* end for(fieldnum) */

	PQclear(result);

	/* format caller id string */
	if (cidname[0]) {
		snprintf(cid, sizeof(cid), "\"%s\" <%s>", cidname, cidnumber);
		ast_data_add_string(response, ASTDATA_ATTR_CALLERID, cid);
	}
	else {
		if (cidnumber[0]) {
			ast_data_add_string(response, ASTDATA_ATTR_CALLERID, cidnumber);
		}
	}

	/* go get the acl if there is one */
	pgsql_getacl(response, poolentry);

	ast_data_clear_error(response);

	poolentry->checkedout = 0;
	ast_mutex_unlock(&poolentry->lock);

	return(response);
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_iax_find_user(ASTDATA_PACKET *request) {
	return(pgsql_iax_find_friend(request, IAX_TYPE_USER));
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_iax_find_peer(ASTDATA_PACKET *request) {
	return(pgsql_iax_find_friend(request, IAX_TYPE_PEER));
}

/***************************************************************************/

static ASTDATA_PACKET *
pgsql_end_of_call(ASTDATA_PACKET *request) {
	ast_log(LOG_NOTICE, "<><><>END OF CALL<><><>\n");
	return(NULL);
}

/***************************************************************************/

static int
my_unload_module(void) {
	ast_cli_unregister(&ast_data_status_cli);
	ast_cli_unregister(&ast_data_debug_sql_cli);
	ast_cli_unregister(&ast_data_no_debug_sql_cli);

	ast_data_unregister_all(CLI_NAME);

	destroy_dbpools();

	return(0);
}

/***************************************************************************/

static int
my_load_module(void) {
	int res = 0;
	struct ast_config *cfg = NULL;
	struct ast_variable *var = NULL;
	struct db_entry *dbptr = NULL;
	ASTDATA_DBINFO *default_dbinfo = NULL;

	/* Create the initial/default db_entry
	 */
	dbptr = calloc(1, sizeof(struct db_entry));
	if (dbptr == NULL) {
		/* Out of memory */
		ast_log(LOG_WARNING, "Out of memory defining default database\n");
		ast_destroy(cfg);
		return(0);
	}

	ast_mutex_init(&dbptr->pool_lock);
	dbptr->initial_dbpool_size = DEFAULT_INIT_POOL_SIZE;
	dbptr->max_dbpool_size = DEFAULT_MAX_POOL_SIZE;

	/* remember which dbinfo struct is the default--for checking what
	 * is allowed query for this module
	 */
	default_dbinfo = &dbptr->dbinfo;
	db_list = dbptr;

	cfg = ast_data_read_handler_flags(CLI_NAME, default_dbinfo);
	if (!cfg) {
		return(0);
	}

	/* now, go get the config section specific to this module */
	var = ast_variable_browse(cfg, CLI_NAME);
	if (!var) {
		/* nothing configured */
		ast_log(LOG_WARNING, "no [%s] section in %s\n",
			CLI_NAME, ASTDATA_CONFIG_FILE);
		ast_destroy(cfg);
		return(0);
	}

	while (var) {
		if (!strcasecmp(var->name, "qtype")) {
			if (strcasecmp(var->value, "default")) {
				/* Not default, so allocate some memory */
				dbptr = calloc(1, sizeof(struct db_entry));
				if (dbptr == NULL) {
					/* Out of memory */
					ast_log(LOG_WARNING,
						"Out of memory defining database for %s, line %d of %s\n",
						var->value, var->lineno, ASTDATA_CONFIG_FILE);
				}
				else {
					ast_mutex_init(&dbptr->pool_lock);
				
					dbptr->initial_dbpool_size = DEFAULT_INIT_POOL_SIZE;
					dbptr->max_dbpool_size = DEFAULT_MAX_POOL_SIZE;
	
					/* Add to chain */
					dbptr->next = db_list;
					db_list = dbptr;
				}
			}
			if (dbptr) {
				ast_data_parse_cfg_qtype(var->value, &dbptr->dbinfo, default_dbinfo);
			}
		} /* end if (check for "qtype=") */

		if (db_list) {
			/* db_list should get set, or change during above if-block
			 */
			if (!strcasecmp(var->name, "dboptions")) {
				strncpy(db_list->dbinfo.options, var->value,
					sizeof(db_list->dbinfo.options)-1);
			} else if (!strcasecmp(var->name, "initpoolsize")) {
				if (sscanf(var->value, "%d", &db_list->initial_dbpool_size) < 1) {
					ast_log(LOG_WARNING,
						"Bad InitPoolSize value, line %d of %s.  Using default %d\n",
						var->lineno, ASTDATA_CONFIG_FILE, DEFAULT_INIT_POOL_SIZE);
				}
			} else if (!strcasecmp(var->name, "maxpoolsize")) {
				if (sscanf(var->value, "%d", &db_list->max_dbpool_size) < 1) {
					ast_log(LOG_WARNING,
						"Bad MaxPoolSize value, line %d of %s.  Using default %d\n",
						var->lineno, ASTDATA_CONFIG_FILE, DEFAULT_MAX_POOL_SIZE);
				}
			} else if (!strcasecmp(var->name, "extensionlookupsql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_EXTENSION_LOOKUP, var->value, 0);
			} else if (!strcasecmp(var->name, "sipfindusersql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_SIP_FIND_USER, var->value, 0);
			} else if (!strcasecmp(var->name, "sipfindpeerbynamesql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_SIP_FIND_PEER, var->value, 0);
			} else if (!strcasecmp(var->name, "sipfindpeerbyipsql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_SIP_FIND_PEER, var->value, 1);
			} else if (!strcasecmp(var->name, "sipupdatepeersql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_SIP_UPDATE_PEER, var->value, 0);
			} else if (!strcasecmp(var->name, "vmchangepasswordsql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_VM_CHANGE_PASSWORD, var->value, 0);
			} else if (!strcasecmp(var->name, "vmfindusersql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_VM_FIND_USER, var->value, 0);
			} else if (!strcasecmp(var->name, "vmdirectorylnamesql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_VM_DIRECTORY, var->value, 0);
			} else if (!strcasecmp(var->name, "vmdirectoryfnamesql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_VM_DIRECTORY, var->value, 1);
			} else if (!strcasecmp(var->name, "iaxfindusersql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_IAX_FIND_USER, var->value, 0);
			} else if (!strcasecmp(var->name, "iaxfindpeerbynamesql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_IAX_FIND_PEER, var->value, 0);
			} else if (!strcasecmp(var->name, "iaxfindpeerbyipsql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_IAX_FIND_PEER, var->value, 1);
			} else if (!strcasecmp(var->name, "iaxupdatepeersql")) {
				ast_data_set_qinfo_query(&db_list->dbinfo,
					ASTDATA_QTYPE_IAX_UPDATE_PEER, var->value, 0);
			}
		}
		var = var->next;
	}
	ast_destroy(cfg);

	if (db_list == NULL) {	
		ast_log(LOG_WARNING, "No 'QType' defined in %s\n", ASTDATA_CONFIG_FILE);
		return(0);
	}

	/* Connect to the database */
	if (initialize_dbpools()) return(0);

	/* Enable the handlers per the flags in the config file */

	/* SIP */
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_SIP_FIND_PEER, pgsql_sip_find_peer);
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_SIP_FIND_USER, pgsql_sip_find_user);
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_SIP_UPDATE_PEER, pgsql_sip_update_peer);

	/* IAX */
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_IAX_FIND_PEER, pgsql_iax_find_peer);
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_IAX_FIND_USER, pgsql_iax_find_user);
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_IAX_UPDATE_PEER, pgsql_iax_update_peer);

	/* Extensions */
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_EXTENSION_LOOKUP, pgsql_ext_lookup);

	/* Voicemail */
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_VM_FIND_USER, pgsql_vm_find_user);
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_VM_CHANGE_PASSWORD, pgsql_vm_change_password);
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_VM_DIRECTORY, pgsql_vm_directory);

	/* End of Call */
	ast_data_register(CLI_NAME, TECH_NAME, default_dbinfo,
		ASTDATA_QTYPE_END_OF_CALL, pgsql_end_of_call);
	
	ast_cli_register(&ast_data_status_cli);
	ast_cli_register(&ast_data_debug_sql_cli);
	ast_cli_register(&ast_data_no_debug_sql_cli);

	return(res);
}

/***************************************************************************/

char *
description(void) {
	return(desc);
}

/***************************************************************************/

int
load_module(void) {
	return(my_load_module());
}

/***************************************************************************/

int
unload_module(void) {
	return(my_unload_module());
}

/***************************************************************************/

int
reload(void) {
	my_unload_module();
	return(my_load_module());
}

/***************************************************************************/

int
usecount(void) {
	struct db_entry *dbentry = NULL;
	struct db_pool_entry *poolentry = NULL;
	int count = 0;

	/* count number of pool connections checked out */
	dbentry = db_list;
	while (dbentry) {
		poolentry = dbentry->pool;
		while (poolentry) {
			count += poolentry->checkedout;
			poolentry = poolentry->next;
		}
		dbentry = dbentry->next;
	}
	return(count);
}

/***************************************************************************/

char *
key() {
	return(ASTERISK_GPL_KEY);
}
