/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Data Resource API
 * 
 * Copyright (C) 2004, Rob Gagnon
 * Rob Gagnon <rgagnon@networkip.net
 *
 * Contributions by:
 *   Steven Sokol <ssokol@sokol-associates.com>
 *
 * based on code originally written by:
 * Copyright (C) 1999, Mark Spencer
 * Mark Spencer <markster@linux-support.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License.
 *
 * Includes code and algorithms from the Zapata library.
 *
 */

#ifndef _ASTDATA_H
#define _ASTDATA_H

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>

/* Extension lookup stuff */
#ifndef HELPER_EXISTS
#define HELPER_EXISTS		0
#define HELPER_SPAWN			1
#define HELPER_EXEC			2
#define HELPER_CANMATCH		3
#define HELPER_MATCHMORE	4
#endif

/* ACL Stuff */
#ifndef AST_SENSE_DENY
#define AST_SENSE_DENY		0
#define AST_SENSE_ALLOW		1
#endif

/* SIP stuff */
#ifndef SIP_TYPE_USER
#define SIP_TYPE_USER      (1 << 0)
#define SIP_TYPE_PEER      (1 << 1)
#define SIP_TYPE_FRIEND    (SIP_TYPE_PEER|SIP_TYPE_USER)
#endif

/* IAX stuff */
#ifndef IAX_TYPE_USER
#define IAX_TYPE_USER		(1 << 0)
#define IAX_TYPE_PEER		(1 << 1)
#define IAX_TYPE_FRIEND		(IAX_TYPE_PEER|IAX_TYPE_USER)
#endif

#ifndef IAX_CAPABILITY_FULLBANDWIDTH
			/* Ethernet, etc */
#define IAX_CAPABILITY_FULLBANDWIDTH	0xFFFF

			/* T1, maybe ISDN */
#define IAX_CAPABILITY_MEDBANDWIDTH		(IAX_CAPABILITY_FULLBANDWIDTH & \
														~AST_FORMAT_SLINEAR & \
														~AST_FORMAT_ULAW & \
														~AST_FORMAT_ALAW)

			/* A modem */
#define IAX_CAPABILITY_LOWBANDWIDTH		(IAX_CAPABILITY_MEDBANDWIDTH & \
														~AST_FORMAT_G726 & \
														~AST_FORMAT_ADPCM)
#define IAX_CAPABILITY_LOWFREE			(IAX_CAPABILITY_LOWBANDWIDTH & \
														~AST_FORMAT_G723_1)
#endif

#define ASTDATA_CONFIG_FILE "data.conf"
#define ASTDATA_LEGACY_SIP_CONFIG "sip.conf"
#define ASTDATA_LEGACY_IAX_CONFIG "iax.conf"
#define ASTDATA_LEGACY_VM_CONFIG "voicemail.conf"

typedef int ASTDATA_ERROR;

/* Data Resource Error Types */
#define ASTDATA_E_OK		 			0
#define ASTDATA_E_NOT_PRESENT		-1
#define ASTDATA_E_OTHER				-2
#define ASTDATA_E_UPDATE_FAILED	-3
#define ASTDATA_E_FIND_FAILED  	-4
#define ASTDATA_E_DELETE_FAILED 	-5
#define ASTDATA_E_NOMEM				-6
#define ASTDATA_E_BAD_PARAM		-7
#define ASTDATA_E_NO_DB_CONN		-8
#define ASTDATA_E_NOT_FOUND		-9

typedef int ASTDATA_QTYPE;

/* Data Resource Query Types */
#define ASTDATA_QTYPE_INVALID					0
#define ASTDATA_QTYPE_EXTENSION_LOOKUP				1
#define ASTDATA_QTYPE_SIP_FIND_USER				2
#define ASTDATA_QTYPE_SIP_FIND_PEER				3
#define ASTDATA_QTYPE_SIP_UPDATE_PEER				4
#define ASTDATA_QTYPE_VM_CHANGE_PASSWORD			5
#define ASTDATA_QTYPE_VM_FIND_USER				6
#define ASTDATA_QTYPE_VM_DIRECTORY				7
#define ASTDATA_QTYPE_IAX_FIND_USER				8
#define ASTDATA_QTYPE_IAX_FIND_PEER				9
#define ASTDATA_QTYPE_IAX_UPDATE_PEER				10
#define ASTDATA_QTYPE_END_OF_CALL				11
#define ASTDATA_QTYPE_SIP_FIND_PEERS_WITH_MAILBOX		12

typedef int ASTDATA_TYPE;

/* attribute data types */
#define ASTDATA_TYPE_INVALID	0
#define ASTDATA_TYPE_STRING	1
#define ASTDATA_TYPE_INTEGER	2
#define ASTDATA_TYPE_IPADDR	3
#define ASTDATA_TYPE_DATE		4

typedef int ASTDATA_ATTR;

/* attribute values */
#define ASTDATA_ATTR_INVALID			0
#define ASTDATA_ATTR_ERR_NUM			10
#define ASTDATA_ATTR_ERR_MSG			11
#define ASTDATA_ATTR_DATASOURCE		20
#define ASTDATA_ATTR_NAME				100
#define ASTDATA_ATTR_USERNAME			101
#define ASTDATA_ATTR_USERTYPE			102
#define ASTDATA_ATTR_MAILBOX			103
#define ASTDATA_ATTR_FULLNAME			104
#define ASTDATA_ATTR_EMAIL				105
#define ASTDATA_ATTR_PAGER				106
#define ASTDATA_ATTR_OPTIONS			107
#define ASTDATA_ATTR_PASSWORD			108
#define ASTDATA_ATTR_MD5PASSWORD		109
#define ASTDATA_ATTR_CONTEXT			110
#define ASTDATA_ATTR_CALLERID			111
#define ASTDATA_ATTR_CALLERID_NUM	112
#define ASTDATA_ATTR_CALLERID_NAME	113
#define ASTDATA_ATTR_EXTENSION		114
#define ASTDATA_ATTR_PRIORITY			115
#define ASTDATA_ATTR_APPLICATION		116
#define ASTDATA_ATTR_ARGUMENTS		117
#define ASTDATA_ATTR_ACCOUNTCODE		118
#define ASTDATA_ATTR_ACLSENSE			119
#define ASTDATA_ATTR_IPADDRESS		120
#define ASTDATA_ATTR_NETMASK			121
#define ASTDATA_ATTR_PORT				122
#define ASTDATA_ATTR_EXPIRY			123
#define ASTDATA_ATTR_REGSECONDS		124
#define ASTDATA_ATTR_NATFLAG			125
#define ASTDATA_ATTR_CANREINVITE		126
#define ASTDATA_ATTR_AMAFLAGS			127
#define ASTDATA_ATTR_DTMFMODE			128
#define ASTDATA_ATTR_CALLINGPRES		129
#define ASTDATA_ATTR_CODECFLAGS		130
#define ASTDATA_ATTR_NEWPASSWORD		131
#define ASTDATA_ATTR_ACLIPADDRESS	132
#define ASTDATA_ATTR_ACLNETMASK		133
#define ASTDATA_ATTR_ACTION			134
#define ASTDATA_ATTR_AUTHFLAGS		135
#define ASTDATA_ATTR_PRIVATEKEY		136
#define ASTDATA_ATTR_PUBLICKEY		137
#define ASTDATA_ATTR_NOTRANSFERFLAG	138
#define ASTDATA_ATTR_USERAGENT		139
#define ASTDATA_ATTR_USERINPUT		140
#define ASTDATA_ATTR_LASTNAMEFLAG	141
#define ASTDATA_ATTR_FULLCONTACT		142
#define ASTDATA_ATTR_LANGUAGE			143

typedef struct ast_data_dict_attr {
	ASTDATA_ATTR	attribute; /* one of the ASTDATA_ATTR_xxxx values */
	ASTDATA_TYPE	type;	/* one of the ASTDATA_TYPE_xxxx values */
	char				*name;
} ASTDATA_DICT_ATTR;

typedef struct value_pair {
	ASTDATA_ATTR		attribute;
	int					lvalue;
	struct in_addr		ipaddr;
	char					*strvalue;
	struct value_pair	*next;
} ASTDATA_PAIR;

typedef struct ast_data_packet {
	ast_mutex_t		lock;
	ASTDATA_QTYPE	qtype;
	ASTDATA_PAIR	*data;
} ASTDATA_PACKET;

typedef struct ast_data_query_info {
	ASTDATA_QTYPE					qtype;
	char								*query;
	char								*altquery;
	struct ast_data_query_info	*next;
} ASTDATA_QINFO;

typedef struct ast_data_db_info {
	char				hostname[81];
	char				dbname[81];
	char				password[81];
	char				user[81];
	int				port;
	char				sock[81];
	char				options[256];
	ASTDATA_QINFO	*qinfo;
} ASTDATA_DBINFO;

/*!
 * \brief read knowledge of setup queries for a given data module
 * \param modname character string for data module name (ie: odbc)
 * \param dbinfo ptr for routine to return the flags
 * \return struct ast_config ptr to open config file load
 *
 * Reads the ASTDATA_CONFIG_FILE and finds which queries are enabled for
 * a given modname in the [global] section of the file.  Returns ptr to
 * the config file so further config file reading can take place.  Adds
 * ASTDATA_QINFO items to supplied ASTDATA_DBINFO struct to indicate allowed
 * queries for the module
 */
extern struct ast_config *
ast_data_read_handler_flags(char *modname, ASTDATA_DBINFO *dbinfo);

/*!
 * \brief Escape a string for SQL
 * \param to char ptr to buffer where escaped string will be placed
 * \param from char ptr of source data
 * \param fromlen the number of characters in from to handle
 * \return number of characters placed in the to buffer
 *
 * Escape a string for inclusion with an SQL statment to protect from SQL
 * injection attack
 */
extern int
ast_data_escape_string(char *to, char *from, int fromlen);

/*!
 * \brief Trim whitespace from the right side of a character string
 * \param s char ptr to source string to trim
 * \return char ptr to same string
 *
 * Changes all \n \r, and spaces on right of string to \0 to effectively
 * trim the string
 */
extern char *
ast_data_rtrim(char *s);

/*!
 * \brief Set the Error information on a packet
 * \param packet Pointer to the packet to set the error information on
 * \param errnum ASTDATA_E_xxxx value to set
 * \param errmsg Character string pointer to text of error information
 *
 * Adds or replaces the existing error number and message in a packet 
 */
extern void
ast_data_set_error(ASTDATA_PACKET *packet, ASTDATA_ERROR errnum, char *errmsg);

/*!
 * \brief Clear the error information on a packet
 * \param packet Pointer to the packet to set the error information on
 *
 * Sets the error number and message to ASTDATA_E_OK, and "OK" respectively
 */
extern void
ast_data_clear_error(ASTDATA_PACKET *packet);

/*!
 * \brief Remove a packet from memory
 * \param packet Pointer to the packet to free
 *
 * Free all memory used by a packet and its contained value pairs
 */
extern void
ast_data_free_packet(ASTDATA_PACKET *packet);

/*!
 * \brief Find first occurance of an attribute in a packet
 * \param packet Pointer to the packet to look through
 * \param attribute ASTDATA_ATTR value to locate in the packet
 * \return Pointer to ASTDATA_PAIR if found, or NULL if not found
 *
 * Find first occurance of a value pair in a packet by the attribute value
 */
extern ASTDATA_PAIR *
ast_data_first_pair(ASTDATA_PACKET *packet, ASTDATA_ATTR attribute);

/*!
 * \brief Find next occurance of an attribute in a packet
 * \param prevpair Pointer to a ASTDATA_PAIR struct
 * \param attribute ASTDATA_ATTR value to locate in the packet
 * \return Pointer to ASTDATA_PAIR if found, or NULL if not found
 *
 * Find next occurance of a value pair in a linked-list of pairs
 * by the attribute value.  The attribute associated with prevpair
 * is /b not included in the search.
 */
extern ASTDATA_PAIR *
ast_data_next_pair(ASTDATA_PAIR *prevpair, ASTDATA_ATTR attribute);

/*!
 * \brief Remove an attribute from a packet
 * \param packet Pointer to a packet to search through
 * \param attribute ASTDATA_ATTR value to remove from the packet
 *
 * Removes all pairs in a packet that have a given attribute value
 */
extern void
ast_data_delete_pair(ASTDATA_PACKET *packet, ASTDATA_ATTR attribute);

/*!
 * \brief Add a value pair to a packet
 * \param packet Pointer to ASTDATA_PACKET to add pair to
 * \param new Pointer to ASTDATA_PAIR that will be added
 *
 * Adds an already allocated ASTDATA_PAIR to a given packet.  Pair is added
 * at the end of the linked list inside the packet to maintain order.
 */
extern void
ast_data_add_pair(ASTDATA_PACKET *packet, ASTDATA_PAIR *new);

/*!
 * \brief Add a value pair to a packet with a given integer value
 * \param packet Pointer to ASTDATA_PACKET to add pair to
 * \param attr ASTDATA_ATTR value to add the integer under
 * \param lvalue int value to assign to the attribute
 * \return Pointer to the ASTDATA_PAIR added to the packet, or NULL if no memory
 *
 * Allocates memory for a ASTDATA_PAIR and adds to a given packet under a
 * given attribute value.
 */
extern ASTDATA_PAIR *
ast_data_add_integer(ASTDATA_PACKET *packet, ASTDATA_ATTR attr, int lvalue);

/*!
 * \brief Add a value pair to a packet with a given date value
 * \param packet Pointer to ASTDATA_PACKET to add pair to
 * \param attr ASTDATA_ATTR value to add the date under
 * \param datevalue date value to assign to the attribute
 * \return Pointer to the ASTDATA_PAIR added to the packet, or NULL if no memory
 *
 * Allocates memory for a ASTDATA_PAIR and adds to a given packet under a
 * given attribute value.
 */
extern ASTDATA_PAIR *
ast_data_add_date(ASTDATA_PACKET *packet, ASTDATA_ATTR attr, int datevalue);

/*!
 * \brief Add a value pair to a packet with a given string value
 * \param packet Pointer to ASTDATA_PACKET to add pair to
 * \param attr ASTDATA_ATTR value to add the string under
 * \param strvalue String value to assign to the attribute
 * \return Pointer to the ASTDATA_PAIR added to the packet, or NULL if no memory
 *
 * Allocates memory for a ASTDATA_PAIR and adds to a given packet under a
 * given attribute value.  The string coming in will be strdup'd to isolate
 * it from the calling routine.
 */
extern ASTDATA_PAIR *
ast_data_add_string(ASTDATA_PACKET *packet, ASTDATA_ATTR attr, char *strvalue);

/*!
 * \brief Add a value pair to a packet with a given IP Address value
 * \param packet Pointer to ASTDATA_PACKET to add pair to
 * \param attr ASTDATA_ATTR value to add the IP Address under
 * \param ipaddr IP Address value to assign to the attribute
 * \return Pointer to the ASTDATA_PAIR added to the packet
 * \return NULL if no memory
 *
 * Allocates memory for a ASTDATA_PAIR and adds to a given packet under a
 * given attribute value.
 */
extern ASTDATA_PAIR *
ast_data_add_ipaddr(ASTDATA_PACKET *packet, ASTDATA_ATTR attr, char *ipaddr);

/*!
 * \brief Check a packet for the existence of a list of Attributes
 * \param packet Pointer to a ASTDATA_PACKET struct to check
 * \param ... A NULL terminated comma-delimited list of ASTDATA_ATTR values
 * \return ASTDATA_E_OK if all attributes listed are present within
 * \return ASTDATA_E_BAD_PARAM if packet is NULL
 * \return ASTDATA_ATTR_xxxx  attribute id of missing attribute
 *
 * Check a packet to see if all of the given attributes are contained in the
 * value pairs associated with the packet.
 */
extern ASTDATA_ATTR
ast_data_check_params(ASTDATA_PACKET *packet, ...);

/*!
 * \brief Extract data from a packet
 * \param packet Pointer to a ASTDATA_PACKET struct to extract from
 * \param ... A NULL terminated comma-delimited list of ASTDATA_ATTR and pointer
 * \return ASTDATA_E_OK if all attributes listed are present within
 * \return ASTDATA_E_BAD_PARAM if packet is NULL
 * \return ASTDATA_ATTR_xxxx  attribute id of missing attribute
 *
 * Extract attribute(s) from a packet, and assign to the pointer following
 * the attribute name.
 */
extern ASTDATA_ATTR
ast_data_extract_packet(ASTDATA_PACKET *packet, ...);

/*!
 * \brief Create a new packet, and assign attribute-value pairs to it
 * \param qtype ASTDATA_QTYPE to assign to the packet
 * \param ... A NULL terminated comma-delimited list of ASTDATA_ATTR and pointer
 * \return Pointer to ASTDATA_PACKET struct created
 * \return NULL if no memory
 *
 * Allocates memory, and adds pair(s) of values to the packet.
 */
extern ASTDATA_PACKET *
ast_data_create_packet(ASTDATA_QTYPE qtype, ...);

/*!
 * \brief Get the error string contained in a packet
 * \param packet Pointer to ASTDATA_PACKET struct to get the error from
 * \return Pointer to character string containing the error message
 * \return NULL if no error message in the packet
 */
extern char *
ast_data_error_message(ASTDATA_PACKET *packet);

/*!
 * \brief Get the error number contained in a packet
 * \param packet Pointer to ASTDATA_PACKET struct to get the error from
 * \return ASTDATA_E_xxxx value from packet if present
 * \return ASTDATA_E_NOT_PRESENT if no error is present in the packet
 */
extern ASTDATA_ERROR
ast_data_error_number(ASTDATA_PACKET *packet);

/*!
 * \brief Print the contents of a packet
 * \param packet Pointer to a ASTDATA_PACKET struct
 *
 * Outputs the contents of a packet to the asterisk logs
 */
extern void
ast_data_dump_packet(ASTDATA_PACKET *packet);

/*!
 * \brief Find the name for an attribute
 * \param attribute
 * \return character string pointer to human readable name of attribute
 *
 * Get the name of a given attribute from the attribute dictionary
 */
extern char *
ast_data_attr_name(ASTDATA_ATTR attribute);

/*!
 * \brief Find attribute for a name
 * \param name
 * \return ASTDATA_ATTR associated with the given name
 *
 * Get the attribute value of a given name from the attribute dictionary
 */
extern ASTDATA_ATTR
ast_data_name2attr(char *name);

/*!
 * \brief Find the data type for an attribute
 * \param attribute
 * \return ASTDATA_TYPE associated with the given attribute
 *
 * Get the data type of a given attribute from the attribute dictionary
 */
extern ASTDATA_TYPE
ast_data_attr_type(ASTDATA_ATTR attribute);

/*!
 * \brief Indicate is a given query type is registered
 * \param qtype a ASTDATA_QTYPE value to check
 * \return -1 if something is fully registered, not just reserved
 * \return 0 otherwise
 *
 * Return a True(-1)/False(0) indicating if any data handling module has
 * fully registered a given ASTDATA_QTYPE query type
 */
extern int
ast_data_isqtype_registered(ASTDATA_QTYPE qtype);

/*!
 * \brief Validate Query Type
 * \param qtype int to validate as a good Data Resource Query Type
 * \return -1 if valid, 0 otherwise
 *
 * Used to validate a query type
 */
extern int
ast_data_isvalid_qtype(ASTDATA_QTYPE qtype);

/*!
 * \brief Convert a query type to String
 * \param qtype int to get the name of
 * \return pointer to character string related to query type
 *
 * Used to get a human-readable version of a query type
 */
extern char *
ast_data_qtype2str(ASTDATA_QTYPE qtype);

typedef ASTDATA_PACKET * (*ast_data_query_handler)(ASTDATA_PACKET *request);

/*!
 * \brief Find a query_info struct present in a dbinfo struct
 * \param dbinfo pointer to the dbinfo struct to search
 * \param qtype the query type to look for
 * \return query info struct if found, or NULL
 *
 * Used to search a single dbinfo struct's list of query info structs for
 * the correct query struct, for locating SQL statement for that query etc.
 */
extern ASTDATA_QINFO *
ast_data_find_qinfo(ASTDATA_DBINFO *dbinfo, ASTDATA_QTYPE qtype);

extern char *
ast_data_set_qinfo_query(ASTDATA_DBINFO *dbinfo, ASTDATA_QTYPE qtype, char *query, int usealtquery);

/*!
 * \brief assign a query with its type to a db_info struct
 * \param dbinfo pointer to the dbinfo struct
 * \param qtype the query type to assign
 * \param query pointer to query string, or NULL if query assigned later
 * \return pointer to the query info struct added, or NULL if problems
 *
 * Used to quickly assign a query type/query string combination to a
 * db_info struct
 */
extern ASTDATA_QINFO *
ast_data_add_qinfo(ASTDATA_DBINFO *dbinfo, ASTDATA_QTYPE qtype, char *query);

/*!
 * \brief assign a query with its type to a db_info struct on condition
 * \param to pointer to the target dbinfo struct
 * \param check pointer to the target to check
 * \param qtype the query type to find in check before adding to dest
 * \param query pointer to query string, or NULL if query assigned later
 * \return pointer to the query info struct added, or NULL if problems
 *
 * Similar to ast_data_add_qinfo, but will only add entry to dest if the qtype
 * exists within check already
 */
extern ASTDATA_QINFO *
ast_data_add_qinfo_conditional(ASTDATA_DBINFO *dest, ASTDATA_DBINFO *check, ASTDATA_QTYPE qtype, char *query);

/*!
 * \brief checks if qtype is not default for database specific query assignment
 * \param qtypename character string value for QType =  from data.conf
 * \param dest pointer to target DBINFO struct to flag query if present in check
 * \param check pointer to DBINFO struct used as master available query list
 * \return ASTDATA_E_OK if nothing went wrong
 * \return ASTDATA_E_OTHER if unknown QType
 *
 * Adds a QINFO struct to the list in the dest DBINFO struct if the QINFO
 * struct is already present inside the check DBINFO struct.  The check DBINFO
 * struct is the default DBINFO struct which 'knows' which queries are allowed
 * on the system
 */
extern ASTDATA_ERROR
ast_data_parse_cfg_qtype(char *qtypename, ASTDATA_DBINFO *dest, ASTDATA_DBINFO *check);

/*!
 * \brief free memory held by list of query_info structs in a db_info struct
 * \param dbinfo pointer to the dbinfo struct 
 * \return ASTDATA_E_OK
 *
 * Walk the list of struct query_info's within a struct db_info and free 
 * memory for each.
 */
extern ASTDATA_ERROR
ast_data_free_qinfo(ASTDATA_DBINFO *dbinfo);

ASTDATA_ERROR
ast_data_make_reservations(ASTDATA_QTYPE qtype, char *qorder);

ASTDATA_ERROR
ast_data_reserve(char *name, ASTDATA_QTYPE qtype);

/*!
 * \brief Register a Query Handler
 * \param cliname name associated with the particular Query Handler
 * \param techname description of the Query Handler Technology
 * \param dbinfo ASTDATA_DBINFO struct to check for qtype within
 * \param qtype A query interface query type
 * \param handler function pointer to a Query Handler
 * \return ASTDATA_E_OTHER on error, ASTDATA_E_OK on success.
 *
 * Used to register a Query handler.
 */
extern ASTDATA_ERROR
ast_data_register(char *cliname, char *techname, ASTDATA_DBINFO *dbinfo, ASTDATA_QTYPE qtype, ast_data_query_handler handler);

/*!
 * \brief Unregister a Query handler
 * \param name name of Query handler to unregister
 * \param qtype A query interface query type.  If 0, unregisters all handlers for the handler
 *
 * Unregisters a Query Handler by it's name and type
 */
extern void
ast_data_unregister(char *name, ASTDATA_QTYPE qtype);

extern void
ast_data_unregister_all(char *name);

extern ASTDATA_PACKET *
ast_data_do_query(ASTDATA_PACKET *request);

int
ast_data_load(void);

int
ast_data_unload(void);

int
ast_data_reload(void);

/*!
 * \brief Convert a digit to a regular expression string representing letters
 * \param digit char (a single DTMF digit in the range '0' to '9')
 * \return char pointer to regexp string
 *
 * Used to convert a single digit to a corresponding regular expression string
 * representing the letters possible on the key of a standard telephone
 * keypard.  IE:  '2' returns "(a|b|c|A|B|C)"
 * Wrapped by ast_data_dtmf2regexp()
 */
extern char *
ast_data_digit2regexp(char digit);

/*!
 * \brief Convert a string of DTMF digits to a regular expression
 * \param buff char pointer for the resulting string to be placed int
 * \param bufflen int size of buff above
 * \param dtmf char pointer to string of digits to create expression from
 * \return int length of buff created
 *
 * Used to convert a string of DTMF digits to a corresponding regular
 * expression string representing the letters possible, in the correct order
 * from a standard keypard
 * keypard.  IE:  "234" returns "(a|b|c|A|B|C)(d|e|f|D|E|F)(g|h|i|G|H|I)"
 */
extern int
ast_data_dtmf2regexp(char *buff, int bufflen, char *dtmf);

/*!
 * \brief Search and replace a string within another string
 * \param dest char * buffer to store results within
 * \param destsize int the size of dest in bytes
 * \param src char * the source string containing the pattern 'find'
 * \param find char * the string to find within 'src'
 * \param replace char * the string to replace in 'src'
 * \return pointer to dest string
 *
 * Used to find all occurances of 'find' in the string 'src' and replace 
 * them with the string 'replace'.  Result is placed in 'dest' up to the
 * given number of bytes in 'destsize'.  Search is case-INSENSITIVE
 */
extern char *
ast_data_strrepl(char *dest, int destsize, char *src, char *find, char *replace);

extern char *
ast_data_var_replace(char *dest, int destsize, char *template, ASTDATA_PACKET *pkt);
#endif /* _ASTDATA_H */
