 /*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * MWI Task/Event System
 * 
 * Copyright (C) 2005, Michael Shuler
 * Michael Shuler <mike@bwsys.net>
 *
 * based on code by:
 * Mark Spencer <markster@linux-support.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License.
 *
 */

#include <malloc.h>
#include <string.h>
#include <asterisk/channel.h>
#include <asterisk/lock.h>
#include <asterisk/mwi.h>

/*--- The Global MWI update list ---*/
struct ast_mwi_list
{
        struct mwi *mwis;
	int waitcycles;
        ast_mutex_t lock;
}mwil;


/* MWI Theory Of Operation:
   Since polling entire Xfriends tables from the DB (MySQL, ODBC, etc.) every second from the
   do_monitor() function in chan_sip, iax, etc. would be highly inefficient this improves upon
   Asterisk's polling every entry every second concept to just process mailboxes that really 
   need a NOTIFY message to turn on/off their associated Xfriends message lights. 
   This *should* allow Asterisk to scale considerably better.  Most likely further work could 
   be done to the do_monitor() functions to improve the performance and scalability of Asterisk...
*/

/***************************************************************************/
        
inline void mwi_init(void)
{       
	mwil.mwis = NULL;
	mwil.waitcycles = 0;
        ast_mutex_unlock(&mwil.lock);
}

/***************************************************************************/
        
inline void mwi_setwait(int cycles)
{       
        /* This is used to cause a delay of X do_monitor() cycles before we send out any pending MWI.
	   Why?  Well, because if your phones are dynamic and haven't registered for the first time
           all of your MWI will fail.
	*/

        /* Lock the global list because we are about to modify it */
        ast_mutex_lock(&mwil.lock);
        
	mwil.waitcycles = cycles;
       
	ast_mutex_unlock(&mwil.lock);
}       

/***************************************************************************/
        
inline int mwi_getanddecwait(void)
{       
        /* This is used to cause a delay of X do_monitor() cycles before we send out any pending MWI.
	   Why?  Well, because if your phones are dynamic and haven't registered for the first time
           all of your MWI will fail.
	*/
	/* This counts down one cycle */

        /* Lock the global list because we are about to modify it */
        ast_mutex_lock(&mwil.lock);
	
	if(mwil.waitcycles != 0)
		mwil.waitcycles--;

        ast_mutex_unlock(&mwil.lock);
	return(mwil.waitcycles);
}       
        
/***************************************************************************/
        
inline struct mwi *mwi_gettask(void)
{       
        /* This will return the next mailbox that needs a MWI sent out to all associated Xfriends */
        return(mwil.mwis);
}       
        
/***************************************************************************/

inline int mwi_addtask(char *mailbox, int newmsgs, int oldmsgs)
{
        struct mwi *newmwi;

        /* This will add a mailbox that needs to be checked against all Xfriends tables for who needs to be updated. */

        /* Allocate memory for the new MWI entry */
        if((newmwi = (struct mwi *) calloc(1, sizeof(struct mwi))) == NULL)
                return(-1);

        /* Copy the new mailbox name */
        strcpy(newmwi->mailbox, mailbox);

	/* Set the message count */
	newmwi->newmsgs = newmsgs;
	newmwi->oldmsgs = oldmsgs;

        /* Lock the global list because we are about to modify it */
        ast_mutex_lock(&mwil.lock);

        /* Setup who will be next after we get inserted into the front of the list */
        newmwi->nextmwi = mwil.mwis;

        /* Insert it into the front of the list */
        mwil.mwis = newmwi;

        /* Release the lock */
        ast_mutex_unlock(&mwil.lock);

        return(1);
}

/***************************************************************************/

inline int mwi_deltask(void)
{
        struct mwi *oldmwi;

        /*  This will delete the current MWI which should have just been used */

        /* Lock the global list */
        ast_mutex_lock(&mwil.lock);

        /* Save the soon to be deleted MWI so it can be freed later */
        oldmwi = mwil.mwis;

        /* Knock it off the front of the list by simply moving the pointers around */
        mwil.mwis = mwil.mwis->nextmwi;

        /* Free up the now unused MWI */
        free(oldmwi);

        /* Release the lock */
        ast_mutex_unlock(&mwil.lock);

        return(1);
}

/***************************************************************************/
